import os

import pandas as pd
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QTableWidgetItem, QApplication
from yangke.base import execute_function_by_interval

from yangke.common.config import logger
from yangke.common.qt import YkWindow, run_app, YkDataTableWidget, YkConsole
from yangke.common.QtImporter import QFileDialog, QMessageBox
from yangke.sis.export_history_values import load_history_file, find_condition
from yangke.sis.dll_file import read_data


class MainFrame(YkWindow):
    def __init__(self):
        super(MainFrame, self).__init__()
        self.setWindowTitle("历史工况查询工具")
        self.enable_input_panel()
        self.enable_table()
        self.console = YkConsole()
        self.add_panel('终端', self.console, Qt.BottomDockWidgetArea)
        logger.add(sink=self.console)  # 日志将打印到GUI界面上

    def choose_file(self):
        file, _ = QFileDialog.getOpenFileName(self, "选择历史数据文件", os.getcwd(), "*.xlsx;;*.*")
        if os.path.exists(file):
            self.statusBar().showMessage("正在加载历史数据文件，请耐心等待！")
            self.df = load_history_file(file)
            self.statusBar().showMessage("就绪")

    def get_condition(self):
        """
        根据输入面板内容获取符合条件的工况

        :return:
        """
        values = self.get_value_of_panel(need_dict=True, need_unit=False)
        unit_condition = [("凝汽器热负荷", float(values.get("凝汽器热负荷")), "1%"),
                          ("环境温度", float(values.get("环境温度")), "±2"),
                          ("环境湿度", float(values.get("环境湿度")), "±10"),
                          ]
        cold_condition = {"循泵方式": values.get("循泵方式"), "机力塔数量": int(values.get("机力塔数量"))}
        auto_loose = values.get("自动放宽条件限制")
        auto_loose = True if auto_loose == "是" else False

        self.res = find_condition(self.df, unit_condition=unit_condition, cold_condition=cold_condition,
                                  auto_loose=auto_loose)
        if self.res is None:
            # 弹窗提示
            QMessageBox.information(self, '提示信息', '历史数据中不存在满足指定条件的工况')
            self.statusBar().showMessage("就绪")
        else:
            self.replace_table(table_ui_file="ui/ui_table.yaml")
            self._table_widget.display_dataframe(self.res)
            self.statusBar().showMessage(f"指定工况下的平均背压为{self.res.mean(numeric_only=True)['当前背压']}")

    def set_points(self):
        self.replace_table("ui/ui_table_set_points.yaml")
        # self.table_widget.set_cell_value(1, 1, "设置导数测点清单")
        # self.table_widget.set_cell_value(2, 1, "测点名")
        # self.table_widget.set_cell_value(2, 2, "标签名")

    def test_points(self):
        _ = os.path.join(os.path.dirname(__file__), "ui", "ui_dcs_test.yaml")
        self.add_content_tab(widget=YkDataTableWidget(from_file=_, root_window=self), tab_name='DCS测点测试')
        self._content_tab.activate_tab('DCS测点测试')

    def _get_value_dcs(self):
        """
        获取DCS测点测试面板上的数据标签名，并更新数据
        """
        table: YkDataTableWidget = self._content_tab.get_tab_panel('DCS测点测试')
        rows = table.rowCount()
        df = table.read_data_in_range(2, rows, 1, 2)
        _ = df.duplicated(subset='测点描述')
        if _.any():
            logger.debug(f"测点描述有重复！")
            return

        _ = df.duplicated(subset='标签名')
        if _.any():
            logger.debug(f"测点标签名有重复")
            return

        if df.shape[0] == 0:
            logger.debug("无待测试的测点！")
            return
        snapshot = read_data(tag_des_read=dict(zip(df['测点描述'], df["标签名"])))
        snapshot.drop(columns="DateTime", inplace=True)
        for col in snapshot.columns:
            _ = table.findItems(col, Qt.MatchExactly)
            if len(_) == 1:
                widget: QTableWidgetItem = _[0]
                x, y = widget.row(), widget.column()
                table.set_cell_value(x, y + 2, snapshot[col].values.item())
            else:
                logger.debug(f"表格中描述可能出现了修改，导致结果无法复制给指定参数！")
        logger.debug("数据更新")

    def _start_get_value_(self, b):
        if b:
            self.info['scheduler'] = execute_function_by_interval(self._get_value_dcs, minute=0, second=10,
                                                                  daemon=False)
        else:
            if self.info.get('scheduler') is not None:
                self.info['scheduler'].shutdown()
                self.info['scheduler'] = None
                logger.debug("测试线程已停止，数据不再更新")

    def export_points_enum_code(self):
        table: YkDataTableWidget = self._content_tab.get_tab_panel('DCS测点测试')
        rows = table.rowCount()
        df = table.read_data_in_range(2, rows, 1, 2)
        lines = []
        for idx, row in df.iterrows():
            line = f'{row["测点描述"]} = "{row["标签名"]}"'
            lines.append(line)
        if len(lines) == 0:
            logger.debug("未检测到测点信息，生成测点信息代码为空")
            return

        _code = """
from enum import Enum, unique
from yangke.base import get_key_value, is_number


@unique
@get_key_value
class TagsRead(Enum):
"""
        _ = "\n\t".join(lines)
        out_code = f"测点信息代码如下（测点描述应为合法的变量标识符，否则导出的代码变量名不合法）：\n{_code}\t{_}"
        logger.info(out_code)
        self._content_tab.activate_tab('终端')


if __name__ == "__main__":
    os.chdir(os.path.dirname(__file__))
    run_app(MainFrame)



import numpy as np
from scipy import spatial


def function_for_TSP(num_points, seed=None):
    if seed:
        np.random.seed(seed=seed)

    points_coordinate = np.random.rand(num_points, 2)  # generate coordinate of points randomly
    distance_matrix = spatial.distance.cdist(points_coordinate, points_coordinate, metric='euclidean')

    # print('distance_matrix is: \n', distance_matrix)

    def cal_total_distance(routine):
        num_points, = routine.shape
        return sum([distance_matrix[routine[i % num_points], routine[(i + 1) % num_points]] for i in range(num_points)])

    return num_points, points_coordinate, distance_matrix, cal_total_distance


def sphere(p):
    # Sphere函数
    out_put = 0
    for i in p:
        out_put += i ** 2
    return out_put


def schaffer(p):
    '''
    二维函数，具有无数个极小值点、强烈的震荡形态。很难找到全局最优值
    在(0,0)处取的最值0
    -10<=x1,x2<=10
    '''
    x1, x2 = p
    x = np.square(x1) + np.square(x2)
    return 0.5 + (np.square(np.sin(np.sqrt(x))) - 0.5) / np.square(1 + 0.001 * x)


def shubert(p):
    '''
    2-dimension
    -10<=x1,x2<=10
    has 760 local minimas, 18 of which are global minimas with -186.7309
    '''
    x, y = p
    part1 = [i * np.cos((i + 1) * x + i) for i in range(1, 6)]
    part2 = [i * np.cos((i + 1) * y + i) for i in range(1, 6)]
    return np.sum(part1) * np.sum(part2)


def griewank(p):
    '''
    存在多个局部最小值点，数目与问题的维度有关。
    此函数是典型的非线性多模态函数，具有广泛的搜索空间，是优化算法很难处理的复杂多模态问题。
    在(0,...,0)处取的全局最小值0
    -600<=xi<=600
    '''
    part1 = [np.square(x) / 4000 for x in p]
    part2 = [np.cos(x / np.sqrt(i + 1)) for i, x in enumerate(p)]
    return np.sum(part1) - np.prod(part2) + 1


def rastrigrin(p):
    '''
    多峰值函数，也是典型的非线性多模态函数
    -5.12<=xi<=5.12
    在范围内有10n个局部最小值，峰形高低起伏不定跳跃。很难找到全局最优
    has a global minimum at x = 0  where f(x) = 0
    '''
    return np.sum([np.square(x) - 10 * np.cos(2 * np.pi * x) + 10 for x in p])


def rosenbrock(p):
    '''
    -2.048<=xi<=2.048
    函数全局最优点在一个平滑、狭长的抛物线山谷内，使算法很难辨别搜索方向，查找最优也变得十分困难
    在(1,...,1)处可以找到极小值0
    :param p:
    :return:
    '''
    n_dim = len(p)
    res = 0
    for i in range(n_dim - 1):
        res += 100 * np.square(np.square(p[i]) - p[i + 1]) + np.square(p[i] - 1)
    return res


if __name__ == '__main__':
    print(sphere((0, 0)))
    print(schaffer((0, 0)))
    print(shubert((-7.08350643, -7.70831395)))
    print(griewank((0, 0, 0)))
    print(rastrigrin((0, 0, 0)))
    print(rosenbrock((1, 1, 1)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2019/8/20
# @Author  : github.com/guofei9987


import numpy as np
from .base import SkoBase
from abc import ABCMeta, abstractmethod
from .operators import crossover, mutation, ranking, selection
from .GA import GeneticAlgorithmBase, GA


class DE(GeneticAlgorithmBase):
    def __init__(self, func, n_dim, F=0.5,
                 size_pop=50, max_iter=200, prob_mut=0.3,
                 lb=-1, ub=1,
                 constraint_eq=tuple(), constraint_ueq=tuple()):
        super().__init__(func, n_dim, size_pop, max_iter, prob_mut,
                         constraint_eq=constraint_eq, constraint_ueq=constraint_ueq)

        self.F = F
        self.V, self.U = None, None
        self.lb, self.ub = np.array(lb) * np.ones(self.n_dim), np.array(ub) * np.ones(self.n_dim)
        self.crtbp()

    def crtbp(self):
        # create the population
        self.X = np.random.uniform(low=self.lb, high=self.ub, size=(self.size_pop, self.n_dim))
        return self.X

    def chrom2x(self, Chrom):
        pass

    def ranking(self):
        pass

    def mutation(self):
        '''
        V[i]=X[r1]+F(X[r2]-X[r3]),
        where i, r1, r2, r3 are randomly generated
        '''
        X = self.X
        # i is not needed,
        # and TODO: r1, r2, r3 should not be equal
        random_idx = np.random.randint(0, self.size_pop, size=(self.size_pop, 3))

        r1, r2, r3 = random_idx[:, 0], random_idx[:, 1], random_idx[:, 2]

        # 这里F用固定值，为了防止早熟，可以换成自适应值
        self.V = X[r1, :] + self.F * (X[r2, :] - X[r3, :])

        # the lower & upper bound still works in mutation
        mask = np.random.uniform(low=self.lb, high=self.ub, size=(self.size_pop, self.n_dim))
        self.V = np.where(self.V < self.lb, mask, self.V)
        self.V = np.where(self.V > self.ub, mask, self.V)
        return self.V

    def crossover(self):
        '''
        if rand < prob_crossover, use V, else use X
        '''
        mask = np.random.rand(self.size_pop, self.n_dim) < self.prob_mut
        self.U = np.where(mask, self.V, self.X)
        return self.U

    def selection(self):
        '''
        greedy selection
        '''
        X = self.X.copy()
        f_X = self.x2y().copy()
        self.X = U = self.U
        f_U = self.x2y()

        self.X = np.where((f_X < f_U).reshape(-1, 1), X, U)
        return self.X

    def run(self, max_iter=None):
        self.max_iter = max_iter or self.max_iter
        for i in range(self.max_iter):
            self.mutation()
            self.crossover()
            self.selection()

            # record the best ones
            generation_best_index = self.Y.argmin()
            self.generation_best_X.append(self.X[generation_best_index, :].copy())
            self.generation_best_Y.append(self.Y[generation_best_index])
            self.all_history_Y.append(self.Y)

        global_best_index = np.array(self.generation_best_Y).argmin()
        self.best_x = self.generation_best_X[global_best_index]
        self.best_y = self.func(np.array([self.best_x]))
        return self.best_x, self.best_y



from yangke.base import AutoSavePara
import os

from yangke.common.config import logger
from yangke.common.qt import YkWindow, run_app, YkItem
from PyQt5.QtWidgets import QFileDialog
from yangke.common.fileOperate import read_csv_ex
from yangke.base import auto_save_para



class MainWindow(YkWindow, AutoSavePara):
    def __init__(self):
        super(MainWindow, self).__init__()
        self.enable_input_panel()
        self.file = self.update_auto_save_para('self.file', None)
        self.set_value_of_panel({"数据文件": self.file})

    def choose_file(self):
        files, _ = QFileDialog.getOpenFileName(self, '选择数据文件', os.getcwd(), "All Files(*)")
        self.set_value_of_panel({"数据文件": files})
        self.file = self.update_auto_save_para('self.file', files)
        print("选择文件")

    def read_data(self):
        files = self.get_value_of_panel(need_unit=False, need_dict=True)['数据文件']
        if isinstance(files, list):
            for file in files:
                if os.path.exists(file):
                    df = read_csv_ex(file)
        else:
            if os.path.exists(files):
                df = read_csv_ex(files)
        self.enable_table()
        self._table_widget.display_dataframe(df)
        items = [
            YkItem(label="选择神经网络数据参数："),
            YkItem(unit=["1", "2"])
        ]
        self._input_panel.append_item(items)

    def append_item_in_input_panel(self):
        items = [
            YkItem(label="选择神经网络输入参数："),
            YkItem(unit=["1", "2"])
        ]
        self._input_panel.append_item(items)



import setuptools
import os
from gettext import gettext as _

# 国际化：生成语言目录，需要确保相应目录已经创建
# python setup.py extract_messages -k _ --output-file locale/main.pot   # 生成模板，会覆盖已经存在的模板
# python setup.py init_catalog --domain main -l zh_CN -i locale/main.pot -d locale  # 生成po文件
# python setup.py compile_catalog --domain main -l zh_CN -d locale  # 等同上一句

# 发布wheel安装文件
# python setup.py sdist bdist_wheel

# ----------------------- 离线安装1 ------------------------------
# 首先，离线安装最好保证离线计算机和在线计算机架构及软件版本尽量一致。
# 1.安装anaconda,
# 2.复制在线安装生成的conda环境至离线计算机
# 3.win32api库需要运行Scripts\pywin32_postinstall.py，运行方式为：
#   a) 切换到指定env环境，conda activate <python38>
#   b) 运行脚本： python .\Scripts\pywin32_postinstall.py -install
# 如果以上方法迁移后，运行python时出现错误：无法启动此程序，因为计算机中丢失api-ms-win-core-path-l1-1-0.dll尝试重新安装该程序以解决此问题。
# 该错误可能是低版本的anaconda安装高版本的python导致的，如离线机器Anaconda3-2021.05上安装python3.9会出现该问题。
# 解决方法：
# 1.在联网计算机上安装python 3.8，命令如下：
#   a) conda create --name=python38 python=3.8
# 2.将安装的python3.8的虚拟环境复制到离线电脑上
# 3.离线电脑试运行python3.8，如果不再报错，则继续安装以下额外第三方库
# 4.pip download yangke[All] torch torchvision torchaudio tensorflow -d D:/whl
# 5.将D:/whl文件夹复制到离线机器上。
# 6.pip install --no-index --find-links=D:/whl yangke[All] torch torchvision torchaudio tensorflow
# ----------------------- 离线安装1 ------------------------------

# ----------------------- 离线安装2 ------------------------------
# 1.在离线计算机中安装python3.9.12
# 2.在联网机器上安装好所有包，开始在联网计算机上进行以下操作
# 3.执行命令：pip freeze > require.txt，将安装包的信息保存到txt文件
# 4.执行命令: pip wheel -w DIR -r require.txt
# 5.执行命令: pip download -d DIR -r require.txt，如果生成的require.txt中yangke库是本地路径，则删除后面的路径
# 6.将require.txt和dir目录拷贝到离线计算机中，以下为在离线计算机中的操作
# 7.执行命令pip install --no-index --find-links=DIR -r require.txt
# ----------------------- 离线安装2 ------------------------------

# ----------------------- 上传pypi -----------------------------
# twine upload dist/*
# 输入账号密码即可
# yangke08
# YangKe.12
# ----------------------- 上传pypi -----------------------------

from yangke import __version__, extras_require

yangke_version = __version__
with open(os.path.join(os.path.dirname(__file__), 'README.md'), "r", encoding="utf8") as fh:
    long_description = fh.read()

install_requires = [  # 申明依赖包，安装包时pip会自动安装
    'pandas>=1.3.0',
    'pyyaml>=5.2',
    'json5>=0.8.5',
    'dill>=0.3.3',  # 储存函数对象
    'babel>=2.9',
    'pathlib2',
    'pillow>=7.0.0',
    # 'pywin32',  # 如果添加了pywin32依赖，则无法在linux系统上安装，因为依赖无法满足
    'cmake',
    'loguru',
    'apscheduler',
    'matplotlib>=3.5.0',
    'openpyxl>3',
    'pyKriging',
    'opencv-python',
    'psutil',
    'markupsafe>=2.1.1',  # scrapy框架需要
    'pyecharts',
    'iapws>=1.5',
    'qt_material',

],

all_lib = []
for v in extras_require.values():
    all_lib.extend(v)
all_lib = list(set(all_lib))
extras_require['All'] = all_lib
extras_require['ugly'] = [  # 容易出错的包
    'dlib>=19.17.0',  # dlib需要单独安装，涉及到cmake和boost
    'torch>=1.4.0',  # pytorch需要单独安装，pypi里版本太老
    'scikit_learn',  # 32位python中该库无法从源代码编译安装，需要下载whl文件安装
    'scipy',  # 32位无法从源代码编译安装，需要下载whl文件安装
    'PyQt6-WebEngine',  # 32位python无法安装该库
    'pyWinhook>=1.6.2',  # 32位python无法编译安装，需要下载whl文件安装
    'scikit-image>=0.19.3',  # paddleocr依赖
    'lmdb>=1.3.0',  # paddleocr依赖
    'orjson>3.7.3',  # paddleocr依赖
]

setuptools.setup(
    name="yangke",  # 模块名
    version=yangke_version,
    packages=setuptools.find_packages(
        exclude=['StockData', 'yangke..idea', 'yangke..git']),  # 这里输入Mod包文件目录
    # scripts=["sis_io.py"],
    py_modules=["main",
                # "yangke/spider/stock10jqka/start_stock10jqka",
                # "yangke\spider\stock10jqka\stock10jqka\spiders\jqka_spider",
                ],  # 打包的*.py文件

    package_data={
        "": ["*.py", "*.cfg", "*.js", "*.html", "*.ico"],
    },
    include_package_data=True,  # 将MANIFEST.in中的文件也打包进wheel文件中

    python_requires=">=3.6",

    # metadata to display on PyPI
    author="杨可",
    author_email="yangyangyangkekeke@qq.com",
    description=_("个人工具综合平台，包含常用工具，网络爬虫，知识图谱，神经网络预测等工具"),
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="yangke",
    url="https://gitee.com/yangke02/lib4python",
    project_urls={
        "Bug Tracter": "https://gitee.com/yangke02/lib4python",
        "Documentation": "https://gitee.com/yangke02/lib4python",
        "Source Code": "https://gitee.com/yangke02/lib4python",
        "Funding": "https://gitee.com/yangke02/lib4python",
    },
    classifiers=[  # 给pip工具一些额外的元数据信息，只是用来给pypi提供搜索依据的，对实际项目不做任何限制
        "Development Status :: 4 - Beta",
        # Indicate who your project is intended for
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Build Tools",

        "License :: Freely Distributable",
        "Natural Language :: Chinese (Simplified)",

        "Programming Language :: Python :: 3",
        "Operating System :: Microsoft :: Windows",
    ],
    install_requires=install_requires,
    extras_require=extras_require,
    message_extractors={
        'yangke': [
            ('**.py', 'python', None),
            # ('**/templates/**.html', 'genshi', None),
            # ('**/templates/**.txt', 'genshi', {'template_class': 'genshi.template:TextTemplate'})
        ],
    },
)



# -*- coding: utf-8 -*-
import ctypes
import math
import os.path
import time
from datetime import datetime, timedelta
from enum import Enum, EnumMeta
from typing import Optional
import numpy as np
import pandas as pd

from yangke.base import merge_two_dataframes, get_settings
from yangke.common.config import logger
from yangke.core import runCMD, str_in_list

标煤热值 = 29307.6  # kJ/kg


def ctypes_str2char_array(string: str):
    """
    python3中的字符串是以utf-8格式编码的，为了将字符串传递给C函数，需要将其解码为byte类型，对应C函数中的char*类型或char[]类型

    :param string:
    :return:
    """
    return string.encode("utf8")


def init_write_sis(ip=None, user=None, passwd_str=None, port=None):
    """
    使用dbp_api.write_snapshot_by_cmd()方法时需要先调用该方法设置相关SIS操作脚本的路径

    :return:
    """
    import sys
    import yangke.common.fileOperate as fo
    path = sys.executable
    write_sis_bat = os.path.join(os.path.dirname(__file__), "write_sis.bat")
    write_sis_py = os.path.join(os.path.dirname(__file__), "write_sis.py")
    temp_write_sis_bat = os.path.join(os.path.dirname(__file__), "temp_write_sis.bat")
    temp_write_sis_py = os.path.join(os.path.dirname(__file__), "temp_write_sis.py")
    lines = fo.read_lines(write_sis_bat)
    lines_new = []
    for line in lines:
        line = line.replace("%python_exec%", path)
        line = line.replace("%py_file%", os.path.abspath(temp_write_sis_py))
        lines_new.append(line)
    fo.write_lines(temp_write_sis_bat, lines_new)
    if ip is not None and user is not None and passwd_str is not None and port is not None:
        lines = fo.read_lines(write_sis_py)
        lines_new = []
        for line in lines:
            line = line.replace('@ip@', ip)
            line = line.replace('@user@', user)
            line = line.replace('@passwd@', passwd_str)
            line = line.replace('@port@', str(port))
            lines_new.append(line)

        fo.write_lines(temp_write_sis_py, lines_new)


def init_dbp_api(settings=None, ip=None, port=None, user=None, password=None):
    """
    初始化RDBP代理服务器连接，以settings中的配置为优先，如果需要覆盖代理配置，则可以手动传入settings={}

    :param settings:
    :param ip:
    :param port:
    :param user:
    :param password:
    :return:
    """
    from yangke.common.config import logger

    if settings is None:
        settings = get_settings()
    _ip = settings.get_settings("sis.ip")
    _ip = None if _ip == {} else _ip
    _port = settings.get_settings("sis.port")
    _port = None if _port == {} else _port
    _user = settings.get_settings("sis.user")
    _user = None if _user == {} else _user
    _password = settings.get_settings("sis.password")
    _password = None if _password == {} else _user

    ip = _ip if _ip is not None else ip
    port = _port if _port is not None else port
    user = _user if _user is not None else user
    password = _password if _password is not None else password

    try:
        dbp_api = DllMode(ip, user, password, port)
        return dbp_api
    except:
        logger.warning("RDB代理服务器连接失败")
        return None


def read_data_from_dcs(tag_des_read, url):
    """
    使用控制公司开放API读取数据
    访问地址：http://172.18.248.80:8080/nodiot/restful/redisValue


    :return:
    """
    import requests
    if not isinstance(tag_des_read, dict):
        描述 = tag_des_read.get_keys()
        标签 = tag_des_read.get_values()
        data = {
            "tagList": "CF159FD40MAG20FP002_XH54"
        }
        x = requests.post(url, data=data)
    return x


def read_data(tag_des_read, need_detail=False, ip=None, port=None, user=None, password=None):
    """
    读取标签数据，返回以描述为表头的dataframe对象，从settings.yaml中加载代理服务器信息

    当tag_des_read为字典时示例如下：
    tag_heat_supply_read = {"N1DCS.TOTALMW": "#1电功率", "N2DCS.TOTALMW": "#2电功率", "N3DCS.TOTALMW": "#3电功率"}
    snapshot = read_data(tag_des_read)

    当tag_des_read为Enum对象时示例如下：
    from enum import Enum, unique
    from yangke.base import get_key_value
    @unique
    @get_key_value
    class tag_des_read(Enum):
        电功率1 = "N1DCS.TOTALMW"
        电功率2 = "N2DCS.TOTALMW"
        电功率3 = "N3DCS.TOTALMW"

    snapshot = read_data(tag_des_read)

    :param password:
    :param user:
    :param port:
    :param ip:
    :param tag_des_read: {tag1: des1, tag2: des2}类型的数据，或@get_key_value修饰的Enum对象
    :param need_detail:
    :return:
    """
    dbp_api = init_dbp_api(ip=ip, port=port, user=user, password=password)
    if dbp_api is not None:
        if not isinstance(tag_des_read, dict):
            tags = tag_des_read.get_values()
            des = tag_des_read.get_keys()
            snapshot = dbp_api.get_snapshot(tags=tags, tag_description=des, need_detail=need_detail)
        else:
            snapshot = dbp_api.get_snapshot(tags=list(tag_des_read.values()),
                                            tag_description=list(tag_des_read.keys()),
                                            need_detail=need_detail)
    else:
        logger.warning("RDBP服务器连接失败")
        snapshot = {}
    return snapshot


class DllMode:
    def __init__(self, ip=None, user=None, passwd_str=None, port=None, dll_file=None):
        self.ip = ip
        self.user = user
        self.passwd = passwd_str
        self.port = port
        if dll_file is None:
            path = os.path.join(os.path.dirname(__file__), "resource/dbpapi_x64.dll")
        else:
            path = dll_file
        try:
            self.handle: Optional[ctypes.c_uint64] = None
            self.dll = ctypes.cdll.LoadLibrary(path)
        except OSError:
            logger.warning(f"找不到指定的动态链接库！请检查路径{path}")
            raise OSError(f"找不到指定的动态链接库！请检查路径{path}")
        if ip is not None and user is not None and passwd_str is not None and port is not None:
            self.connect(ip, user, passwd_str, port)
        else:
            logger.debug(f"RDBProxy连接信息不全！")

    def __del__(self):
        self.close()

    def connect(self, ip, user, passwd_str, port):
        self.dll.DBPCreate2.restype = ctypes.c_uint64
        ip = ctypes_str2char_array(ip)
        user = ctypes_str2char_array(user)
        passwd = ctypes_str2char_array(passwd_str)
        port = int(port)
        self.handle = ctypes.c_uint64(self.dll.DBPCreate2(ip, user, passwd, port, 0))
        if self.handle is None:
            print("连接创建失败")
            return False
        ret = self.dll.DBP_Connect(self.handle)
        if 0 == ret:
            return True
        else:
            print("服务器连接失败")
            return False

    def close(self):
        if self.handle is not None and self.handle.value > 0:
            self.dll.DBP_Close(self.handle)
            self.handle = None

    def is_connect(self):
        ret = self.dll.DBP_IsConnect(self.handle)
        if 0 == ret:
            return True
        return False

    def dis_connect(self):
        """
        断开连接

        :return:
        """
        ret = self.dll.DBP_DisConnect(self.handle)
        if 0 == ret:
            return True
        return False

    def get_his_value(self, tags: list or str, tags_description=None, start_time: datetime = None,
                      end_time: datetime = None,
                      time_interval=10, use_description=True):
        """
        待验证

        :param tags:
        :param tags_description: 参考get_snapshot中的同名参数
        :param start_time: datetime.datetime类型，默认为两小时前的时间
        :param end_time: datetime.datetime类型，默认为当前时间
        :param time_interval: 时间间隔，单位s
        :param use_description: 参考get_snapshot中的同名参数
        :return:
        """
        now = datetime.now()
        start_time = start_time or now - timedelta(days=0, hours=2)  # 默认两小时前的时间为读数起始时间
        end_time = end_time or now - timedelta(days=0, hours=0)  # 默认一小时前的时间为读数结束时间
        start_time_long = int(time.mktime(start_time.timetuple()))  # 将时间转为UNIX时间
        end_time_long = int(time.mktime(end_time.timetuple()))
        start_time_c = ctypes.c_long(start_time_long)  # c_long和c_ulong貌似都可以
        end_time_c = ctypes.c_long(end_time_long)
        insert_time = ctypes.c_long(time_interval)  # 插值时间，只有下方flag为1是才有效
        flag = ctypes.c_long(1)  # 标记，为0时取样本值，插值时间参数无效，为1时使插值时间参数生效
        # 预留的点数，如果读取的时间段内点数超过这个数，结果会被裁剪，如果读取的点数少于这个数，会补零
        data_num = math.ceil((end_time_long - start_time_long) / time_interval)
        a = ctypes.c_int(10)
        b = ctypes.c_int(10)
        value_type = ctypes.pointer(a)  # 指针和数组在python传递时有所区别
        data_size_actual = ctypes.pointer(b)  # 返回实际读到的数据个数
        value_double_arr = (ctypes.c_double * data_num)()
        value2_arr = (ctypes.c_long * data_num)()
        time_long_arr = (ctypes.c_long * data_num)()
        qas_short_arr = (ctypes.c_short * data_num)()
        if isinstance(tags, str):  # 读取单个参数的历史数据，tags为数据库标签名
            tag_name = ctypes_str2char_array(tags)  # 名字
            self.dll.DBPGetHisVal(self.handle, tag_name, start_time_c, end_time_c, insert_time, flag,
                                  value_double_arr, value2_arr, time_long_arr, qas_short_arr,
                                  ctypes.c_int(data_num), value_type, data_size_actual)
            return self._assemble_dataframe(tags, time_long_arr, qas_short_arr, value_double_arr, value2_arr,
                                            value_type, None, False)
        else:  # 读取多个数据的历史数据，tags为数据库标签名组成的列表
            df = None
            if tags_description is not None and use_description:
                kks_des = dict(zip(tags, tags_description))
            else:
                kks_des = dict(zip(tags, tags))
            for tag in tags:
                tag_name = kks_des.get(tag)
                tag_c = ctypes_str2char_array(tag)
                self.dll.DBPGetHisVal(self.handle, tag_c, start_time_c, end_time_c, insert_time, flag,
                                      value_double_arr, value2_arr, time_long_arr, qas_short_arr,
                                      ctypes.c_int(data_num), value_type, data_size_actual)
                _ = self._assemble_dataframe(tag_name, time_long_arr, qas_short_arr, value_double_arr, value2_arr,
                                             value_type, None, False)
                if df is None:
                    df = _
                else:
                    df = merge_two_dataframes(df, _)[0]
            return df

    def get_snapshot(self, tags, tag_description=None, need_detail=False, use_description=True):
        """
        获取给定标签列表的快照数据

        :param tags: 标签名
        :param need_detail:是否需要数据的详细信息，默认不需要，如果为True,则会返回数据质量、错误码等详细信息
        :param tag_description: 标签点的描述
        :param use_description: 当给定点描述时，数据列的标题是否使用点描述代替标签名
        :return:
        """

        n_size = len(tags)
        tag_names = (ctypes.c_char_p * n_size)()  # 名字
        for i in range(n_size):
            tag_names[i] = ctypes_str2char_array(tags[i])

        time_long_arr = (ctypes.c_uint32 * n_size)()  # 时间，系统里的时间应该比当前时间早8小时

        qas_short_arr = (ctypes.c_short * n_size)()  # 质量

        value_double_arr = (ctypes.c_double * n_size)()  # 浮点数类型的值
        value2_arr = (ctypes.c_int32 * n_size)()  # 整形类型的值
        value2_type = (ctypes.c_int32 * n_size)()  # 数据类型
        error_code_arr = (ctypes.c_short * n_size)()  # 数据错误码
        self.dll.DBPGetSnapshot(
            self.handle,  # 句柄
            tag_names,  # char* sTagNames[],  //in,标签名字符串指针数组  //apistring
            time_long_arr,  # long ltimes[],   //in, 时标
            qas_short_arr,  # short snqas[],   //in, 质量
            value_double_arr,  # double  dblvals[],   //in, 存放double值,DT_FLOAT32,DT_FLOAT64存放区
            value2_arr,  # long lvals[],   //in, 存放Long值,DT_DIGITAL,DT_INT32,DT_INT64存放区
            value2_type,  # int  ntypes[],   //in, 数据类型,DT_INT32,DT_FLOAT32等。
            error_code_arr,  # short errs[],    //in/out, 错误码
            n_size  # int  nsize    //in, 个数
        )
        if tag_description is not None and use_description:  # 如果使用描述，且描述不为空
            return self._assemble_dataframe(tag_description, time_long_arr,
                                            qas_short_arr, value_double_arr, value2_arr,
                                            value2_type,
                                            error_code_arr, need_detail=need_detail)
        else:
            return self._assemble_dataframe(tags, time_long_arr, qas_short_arr,
                                            value_double_arr, value2_arr,
                                            value2_type,
                                            error_code_arr, need_detail=need_detail)

    @staticmethod
    def _assemble_dataframe(tags, time_long_arr, qas_short_arr, value_double_arr, value2_arr, value2_type,
                            error_code_arr, need_detail=False):
        """
        将代理服务器返回的数据组装成dataframe格式的对象

        :param tags: 数据标签
        :param time_long_arr:
        :param qas_short_arr:
        :param value_double_arr:
        :param value2_arr:
        :param value2_type:
        :param error_code_arr:
        :param need_detail: 是否需要数据的详细信息，默认不需要，如果为True,则会返回数据质量、错误码等详细信息
        :return:
        """
        n_size = len(time_long_arr)  # 标签个数
        if not need_detail:
            columns = ["DateTime"]
            if isinstance(tags, str):  # 说明是读历史数据，只有一个变量标签，但有多个时间标签
                columns = ["DateTime", tags]
                data_list = []
                for i in range(n_size):
                    time_str = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(time_long_arr[i]))
                    if value2_type[0] == 3:
                        data_list.append([time_str, value_double_arr[i]])
                    else:
                        data_list.append([time_str, value2_arr[i]])
                result = pd.DataFrame(columns=columns, data=data_list)
            else:
                columns.extend(tags)
                time_str = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(time_long_arr[0]))
                data_list = [time_str]
                for i in range(n_size):
                    if value2_type[i] == 3:  # 如果数据类型==3，则说明读到的是double类型数据
                        data_list.append(value_double_arr[i])
                    else:
                        data_list.append(value2_arr[i])
                result = pd.DataFrame(columns=columns, data=[data_list])
        else:
            result = {}
            for i in range(n_size):
                if isinstance(tags, str):
                    ...
                else:
                    tag = tags[i]
                    columns = ["DateTime", "值", "质量", "数据类型", "错误码"]
                    time_str = time.strftime('%Y-%m-%d %H:%M:%S', time.gmtime(time_long_arr[i]))
                    qas = qas_short_arr[i]
                    value_type = value2_type[i]
                    if value_type == 3:
                        value = value_double_arr[i]
                    else:
                        value = value2_arr[i]
                    error_code = error_code_arr[i]
                    data_list = [time_str, value, qas, value_type, error_code]
                    dataframe = pd.DataFrame(columns=columns, data=[data_list])
                    result.update({tag: dataframe})
        return result

    def write_snapshot_by_cmd(self, tags, values):
        """
        该方法功能和self.write_snapshot_double()完全相同，但是将写入操作重启一个进程进行，这样写入操作报错时，不会影响主程序崩溃。
        如果出现“Python已停止工作”错误，请检查RDBProxy代理和数据库中是否存在相关的测点。

        :param tags:
        :param values:
        :return:
        """
        from yangke.common.config import logger

        for k in tags:
            if k is None:
                logger.error("标签列表中存在空标签，请检查测点名")
                logger.error("写入SIS失败！")
                return None
        tags_values = {k: v for k, v in zip(tags, values)}
        path = os.path.join(os.path.dirname(__file__), "temp_write_sis.bat")
        if not os.path.exists(path):
            if self.ip is not None and self.user is not None and self.passwd is not None and self.port is not None:
                init_write_sis(ip=self.ip, user=self.user, passwd_str=self.passwd, port=self.port)
            else:
                settings = get_settings()  # 说明没有初始化写SIS脚本文件，这里初始化一下
                _ip = settings.get_settings("sis.ip")
                _port = settings.get_settings("sis.port")
                _user = settings.get_settings("sis.user")
                _password = settings.get_settings("sis.password")
                if _ip is not None and _port is not None and _user is not None and _password is not None:
                    init_write_sis(ip=_ip, user=_user, passwd_str=_password, port=_port)
                else:
                    logger.error("请先初始化写快照脚本，init_write_sis(ip, user, port, passwd)")
                    exit()

        cmd = f'"{path}" "{tags_values}"'
        runCMD(command=cmd, wait_for_result=False, output_type="REALTIME_NORETURN")

    def write_snapshot_double(self, tags, values):
        """
        写double类型数据到数据库，该方法可能会导致程序异常退出，建议使用第三方exe独立调用该接口。
        如果出现“Python已停止工作”错误，请检查RDBProxy代理和数据库中是否存在相关的测点。

        :param tags: 标签名列表
        :param values: 数值列表
        :return:
        """
        n_size = len(tags)
        tag_names = (ctypes.c_char_p * n_size)()  # 名字
        time_long_arr = (ctypes.c_uint32 * n_size)()  # 时间
        qas_short_array = (ctypes.c_short * n_size)()  # 质量
        value_double_arr = (ctypes.c_double * n_size)()  # 浮点数类型的值
        value2_arr = (ctypes.c_int32 * n_size)()  # 整形类型的值
        value_type = (ctypes.c_int32 * n_size)()  # 数据类型
        time_long = int(time.time())  # 保证写入的数据点都具有同一个时标
        for i in range(n_size):
            tag_names[i] = ctypes_str2char_array(tags[i])
            time_long_arr[i] = time_long
            qas_short_array[i] = 0
            value_double_arr[i] = values[i]
            value2_arr[i] = 0
            value_type[i] = 3  # 3表示通过value_double_arr传输数据，其他表示通过value2_arr传输数据

        error_code_arr = (ctypes.c_short * 2)()  # 数据错误码，输出信息
        try:
            self.dll.DBPWriteSnapshot(
                self.handle,  # 句柄
                tag_names,  # char* sTagNames[],  //in,标签名字符串指针数组  //apistring
                time_long_arr,  # long ltimes[],   //in, 时标
                qas_short_array,  # short snqas[],   //in, 质量
                value_double_arr,  # double  dblvals[],   //in, 存放double值,DT_FLOAT32,DT_FLOAT64存放区
                value2_arr,  # long lvals[],   //in, 存放Long值,DT_DIGITAL,DT_INT32,DT_INT64存放区
                value_type,  # int  ntypes[],   //in, 数据类型,DT_INT32,DT_FLOAT32等。
                error_code_arr,  # short errs[],    //in/out, 错误码
                n_size  # int  nsize    //in, 个数
            )
        except:
            pass


def get_tag_value(snapshot, tag_description, optional_value=0):
    """
    根据标签描述获取快照中的数据

    :param snapshot:
    :param tag_description:
    :param optional_value: 标签不存在时的替代值，也就是默认值
    :return:
    """
    if isinstance(tag_description, Enum):  # 枚举类的对象
        tag_description = tag_description.name

    if isinstance(snapshot, dict):
        result = snapshot.get(tag_description)
    else:
        try:
            result = snapshot[tag_description][0]
            if isinstance(result, np.int64):
                result = None
            else:
                result = float(snapshot[tag_description][0])
        except KeyError:
            print(f"快照中不包括名为[{tag_description}]的变量，返回None")
            result = None
    if result is None:
        result = optional_value
    return result


class ColdResult:
    def __init__(self):
        """
        冷端优化的结果类，使用方法实力：
        # 初始化设置，只需执行一次
        cold_result = ColdResult()
        cold_result.设置允许的运行参数波动范围(allowed_range)
        cold_result.设置背压标签(["背压1", "背压2"])
        cold_result.设置冷端功率标签(["冷端总功率1", "冷端总功率2"])
        cold_result.设置状态参数标签()

        # 每次优化结果更新时
        cold_result.pass_result(result_opt)
        cold_result.平抑波动()
        result_opt = cold_result.opt_result

        其中，pass_result方法中传入的参数一般应包括：循泵、风机等最佳运行方式，当前实际的发电机功率、背压、环境温度、湿度、大气压力等
        """
        self.allowed_range = None  # 允许的参数波动范围，当机组运行状态在该范围内波动时，冷端优化结果不更新，该值示例如下，只与对象有关
        # [("电功率1", "±2"), ("供热流量", "±4"),("环境温度", "±4"), ("大气压力", "±10")]
        self.allowed_range_dict = {}  # {"电功率": [34, 38], "供热流量": [54, 62], ...}该值是变量，随时间变化
        self.last_state = None
        self.now_state = {}  # 当前时刻机组的实际运行参数
        self.opt_result = {}  # 当前时刻机组的优化后冷端优化结果
        self.bp_label = []  # 背压参数在数据集中的列标签名，一般就是"背压"，但对于全厂优化，可能有多个背压参数，以列表方式传入
        self.state_labels = []  # 状态参数在数据集中的列表签名，一般为["电功率", "供热量", "环境温度"]
        self.p_cold_label = None  # 冷端总功率在数据集中的列标签名
        self.last_opt_result = None  # 最后一个状态中的优化结果值
        self.last_true_state = None  # 最后一个状态中的实际状态值

    def 设置允许的运行参数波动范围(self, range=[]):  # todo
        """
        设置允许的参数波动范围，示例：
        cold_result = ColdResult()
        allowed_range = [("电功率1", "±2"), ("供热流量", "±4"),("环境温度", "±4"), ("大气压力", "±10")]
        cold_result.设置允许的运行参数波动范围(allowed_range)
        allowed_range中的键必须存在于self.pass_result()方法的true_state字典中。即allowed_range中有“电功率1”，则true_state参数也
        必须有"电功率1"。

        :param range:
        :return:
        """
        self.allowed_range = range

    def 设置背压标签(self, label):
        if isinstance(label, list):
            self.bp_label = label
        else:
            self.bp_label = [label]

    def 设置冷端功率标签(self, label):
        self.p_cold_label = label

    def 设置状态参数标签(self, labels):
        if isinstance(labels, list):
            self.state_labels = labels
        else:
            self.state_labels = [labels]

    def pass_result(self, opt_result=None, true_state=None):
        """
        传入新的冷端优化结果，并根据冷端功率大小调节背压大小以符合客观规律。如果某台机组停机，可以传入true_state="stop"或
        true_state={"1": "stop", "2": "run"}

        :param opt_result: 当前时刻的原始优化结果
        :param true_state: 当前时刻的运行值，至少包含电功率、供热、环境以及冷端运行方式等参数
        :return:
        """
        if true_state is None:
            true_state = {}
        if opt_result is None:
            opt_result = {}
        if isinstance(true_state, str):
            if true_state == "stop":
                # 说明是停机状态
                self.last_opt_result = {}
                self.last_true_state = {}
        else:
            group_nums = list(true_state.keys())
            for _gn in group_nums:
                if true_state.get(_gn) == "stop":
                    # 停机时，将_gn号机组的所有参数设置为0
                    ks = list(opt_result.keys())
                    for k in ks:
                        if k.endswith(str(_gn)):
                            self.last_opt_result.update({k: 0})

                    ks = list(true_state.keys())
                    for k in ks:
                        if k.endswith(str(_gn)):
                            self.last_true_state.update({k: 0})
                else:
                    # 将运行的机组的数据设置为传入的数据
                    self.update_group_data(group_num=_gn, opt_result=opt_result, true_state=true_state)
                self.now_state = true_state  # 当前时刻的运行值
                self.opt_result = opt_result

    def update_group_data(self, group_num, opt_result: dict = None, true_state: dict = None):
        """
        更新优化结果和真实结果，只更新指定的机组，机组号必须是标签名的最后一位，如power1表示1号机组的功率。

        :param group_num:
        :param opt_result:
        :param true_state:
        :return:
        """
        ks = list(opt_result.keys())
        for k in ks:
            if k.endswith(str(group_num)):
                self.last_opt_result.update({k: opt_result[k]})

        ks = list(true_state.keys())
        for k in ks:
            if k.endswith(str(group_num)):
                self.last_true_state.update({k: true_state[k]})

    def 背压合理化调整(self):
        def cal_dp(凝汽器热负荷, 冷端电功率变化, 环境温度):
            """
            计算背压变化值，背压变化与机组负荷、供热流量以及冷端设备运行方式有关

            :param power:
            :param heat:
            :param dpower:
            :return:
            """
            ...

        opt_result = self.opt_result
        true_state = self.now_state
        for bpl in self.bp_label:
            if opt_result[bpl] == 0:  # 如果停机状态
                continue
            else:
                _ = true_state[bpl] * (true_state[self.p_cold_label] / opt_result[self.p_cold_label])
                背压差 = abs(_ - true_state[bpl])
            if opt_result[bpl] < true_state[bpl]:
                opt_result[bpl] = true_state[bpl] + 背压差
            else:
                opt_result[bpl] = true_state[bpl] - 背压差
        self.opt_result = opt_result

    # def _get_allowed_range(self):

    def _get_allowed_range_str(self, title):
        for item in self.allowed_range:
            if item[0] == title:
                return item[1]

    def 平抑波动(self, unit=None):
        """
        当机组参数波动较小时，不改变冷端运行方式。

        参数判断分机组进行还是全厂一起进行，当unit值为None时，全厂整体平抑波动，如果需要分机组平抑波动，通过该值传入机组标号，如[1,2]。
        如果是分机组平抑波动，则self.opt_result中变量的最后一个字符需要与机组编号对应，如"power1"表示1号机组的参数。

        :param unit: 参数判断分机组进行还是全厂一起进行
        :return:
        """

        def get_min_max_value(val, range_str):
            # 根据当前值和表示范围的字符串，计算允许的参数的上下限的值
            if range_str.startswith("±"):
                _r = float(range_str[1:])
                min_val = val - _r
                max_val = val + _r
            elif range_str.endswith("%"):
                pass
            return min_val, max_val

        need_update = False  # 是否需要更新画面上的优化结果，如果参数波动小，则不更新，否则，更新
        if unit is None:
            for item in self.allowed_range:
                title, range = item
                if self.last_true_state.get(title) is None:  # 历史数据不存在，则认为满足小幅波动范围，直接更新优化结果为当前计算结果
                    need_update = True
                    break
                if not self.allowed_range_dict[title][0] <= self.now_state[title] <= self.allowed_range_dict[title][1]:
                    need_update = True
                    break

            if need_update:  # 如果需要更新优化结果，更新新的允许的参数波动范围
                for item in self.allowed_range:
                    title, range = item
                    val = self.now_state[title]
                    min_val, max_val = get_min_max_value(val, range)
                    self.allowed_range_dict.update({title: [min_val, max_val]})
            else:  # 如果波动幅度小，则不需要更新优化结果，直接返回上一次优化结果
                self.opt_result = self.last_opt_result
        else:
            for u in unit:
                for item in self.allowed_range:
                    title, _range = item
                    if title.endswith(str(u)):
                        if self.last_true_state.get(title) is None:
                            need_update = True
                            break
                        if not self.allowed_range_dict[title][0] <= self.now_state[title] <= \
                               self.allowed_range_dict[title][1]:
                            need_update = True
                            break
                if need_update:
                    for item in self.allowed_range:
                        title, _range = item
                        if title.endswith(str(u)):  # 只限制运行方式，不限制背压
                            val = self.now_state[title]
                            min_val, max_val = get_min_max_value(val, _range)
                            self.allowed_range_dict.update({title: [min_val, max_val]})
                        # 如果需要更新，则维持现有的opt_result结果不变
                else:
                    for k, v in self.opt_result.items():
                        if k.endswith(str(u)) and not str_in_list(k, self.bp_label):  # 优化后背压仍取实时值
                            self.opt_result.update({k: self.last_opt_result.get(k)})

    def get_result(self):
        return self.opt_result



import copy

from yangke.common.config import logger
from yangke.ebsilon.constant.constant import (Pressure, Temperature, MassFlow, Power, Density, Enthalpy, Unit, Area,
                                              Length, VolumeFlow, SpecificVolume, One, AngleUnit)


class Values:
    def __init__(self, values=None, values_steps=None):
        """
        设备组件的参数对象，维护两个参数字典，其中values字典代表直接的设置值，而result字典代表中间计算数值及计算结果。
        但values和result中的参数可能有重叠，如焓值既可以直接设置，也可以间接计算。

        :param values:
        :param result:
        """
        self.values_steps = values_steps or {0: {}}  # 按迭代步数保存的参数数据，元素是字典类型
        self.values = values or {}  # 字典类型

    def append_values(self, values):
        """
        在迭代部的末尾添加新的数据

        :param values:
        :return:
        """
        self.values = values
        k = len(self.values_steps)
        self.values_steps[k] = copy.deepcopy(self.values)

    def set_values(self, step, values: "dict | Values", set_none=False):
        """
        设置某个迭代步的值

        如果设置的是单个的值，且set_none=False，则相当于是更新某个时间步的某个值，即
        self.set_values(values={"P": Value(1, "MPa")}, step=1)
        和
        self.update(values={"P": Value(1, "MPa")}, set_none=False, step=1)
        效果相同。
        但当set_none为True，这两个方法的效果有区别。set_values会删除values中没有的参数，而update方法不会删除

        :param set_none: 是否用空值覆盖已有值，如果为True，则当前时间步的所有数据都会替换为values，values中没有的参数会被删除
        :param step:
        :param values:
        :return:
        """
        if isinstance(values, Values):
            values = values.get_values_of_step(step=step)
        if step is None:  # 如果未指定时间步，则默认取最后一个已有的时间步
            step = len(self.values_steps) - 1
        if self.values_steps.get(step) is None:  # 如果当前时间步没数据，则初始化为空字典
            self.values_steps[step] = {}

        if set_none:
            self.values_steps[step] = copy.deepcopy(values)  # 无论当前时间步是否有数据，都直接使用传入值覆盖当前值
        else:
            filter_values = self.get_non_null_values(values)
            self.values_steps[step].update(filter_values)

        if step == len(self.values_steps) - 1:  # 如果设置的是最后一个时间步的值，则更新self.values
            self.values = copy.deepcopy(self.values_steps[step])

    def update(self, values: dict, set_none=False, step: int | None = None):
        """
        更新数据，如果step为None，则更新当前迭代步的数据，不会增加新的迭代步，由面板至后台数据更新时，step没有设置
        如果指定了step，则更新指定迭代步的数据

        :param values:
        :param set_none: 是否用空值覆盖已有值，如果为True，则指定参数的值会被values中的空值替换，values中不包含的值不受影响
        :param step: 迭代步
        :return:
        """
        if isinstance(values, Values):
            values = values.get_values_of_step(step=step)
        if step is None:  # 如果未指定时间步，则默认取最后一个已有的时间步
            step = len(self.values_steps) - 1
        if self.values_steps.get(step) is None:  # 如果当前时间步没数据，则初始化为空字典
            self.values_steps[step] = {}

        if set_none:
            self.values_steps[step].update(values)
        else:
            _val = self.get_non_null_values(values)
            self.values_steps[step].update(_val)

        if step == len(self.values_steps) - 1:  # 如果更新的是最后一个时间步的值，则更新self.values
            self.values = copy.deepcopy(self.values_steps[step])

    def clear_result(self):
        # 清除初始状态以后的其他结果
        # 时间步0只保存画面写入的数据，计算结果永远不要写入时间步0
        if len(self.values_steps) > 1:
            self.values_steps = {0: self.values_steps.get(0)}
            self.values = self.values_steps.get(0)

    def _get_symbol(self, _val, need_unit=False):
        """
        获取某个变量的值，根据neet_unit返回相应的值
        """
        if _val.__class__.__name__ in ["float", "int", "str"]:
            if str(_val).strip() != "":  # 如果_val是非空字符串
                res = Value(_val, None) if need_unit else _val
            else:
                res = None  # 没有单位信息，则忽略need_unit参数
        elif isinstance(_val, Value):
            res = _val
        else:  # 其他未知类型，直接返回
            res = _val

        # 至此，res可能是None或Value()
        if need_unit:
            return res
        else:
            if res is None:
                return res
            elif isinstance(res, Value):
                return res.value
            else:
                return res

    def get(self, symbol, need_unit=False):
        """
        获取当前迭代步组件属性的值

        :param symbol: 参数的符号
        :param need_unit: 获取单位为unit的组件属性值
        :return:
        """
        _val = self.values.get(symbol)
        if _val is None:
            return None
        elif _val.__class__.__name__ in ["float", "int", "str"]:
            return Value(_val) if need_unit else _val
        else:
            return _val if need_unit else _val.value

    @staticmethod
    def get_non_null_values(_val):
        if _val is None:
            return {}
        filter_values = {}
        for k, v in _val.items():
            if isinstance(v, Value):
                if str(v.value).strip() != "":
                    filter_values.update({k: v})
            else:
                if v is None:
                    ...
                elif str(v).strip() != "":
                    filter_values.update({k: v})
        return filter_values

    def get_initial(self, symbol=None, need_unit=False, with_none=True):
        """
        获取初始状态下的参数值

        :param symbol: 获取指定的参数
        :param need_unit: 是否返回单位
        :param with_none: 是否返回空值
        :return:
        """
        if symbol is None:  # 获取全部参数时，忽略need_unit参数
            _val = self.values_steps.get(0)
            if with_none:  # 如果需要返回空值
                res = _val
            else:  # 不需要返回空值，则逐个判断并过滤空值
                res = self.get_non_null_values(_val)
            return res
        elif len(self.values_steps) > 0:
            _val = self.values_steps.get(0).get(symbol)
            res = self._get_symbol(_val, need_unit)
            return res
        else:
            return None

    def get_values_of_step(self, step, symbol=None, need_unit=False, with_none=True):
        """
        获取指定时间步的参数值

        :param step: 时间步
        :param symbol: 获取指定的参数
        :param need_unit: 是否返回单位
        :param with_none: 是否返回空值
        :return:
        """
        if symbol is None:  # 获取全部参数时，忽略need_unit参数
            _val = self.values_steps.get(step)
            if with_none:  # 如果需要返回空值
                res = _val
            else:  # 不需要返回空值，则逐个判断并过滤空值
                res = self.get_non_null_values(_val)
            return res
        else:
            _vals = self.values_steps.get(step)
            if _vals is None:
                return None
            else:
                _val = _vals.get(symbol)
            res = self._get_symbol(_val, need_unit)
            return res

    def get_results(self, symbol=None, need_unit=False):
        """
        获取数据中的结果数据

        :param symbol:
        :param need_unit:
        :return:
        """
        if symbol is None:  # 获取全部参数时，忽略need_unit参数
            if len(self.values_steps) > 0:
                _val = self.values_steps[len(self.values_steps) - 1]
                return _val
        else:
            if len(self.values_steps) > 0:
                _val = self.values_steps[len(self.values_steps) - 1].get(symbol)
                res = self._get_symbol(_val, symbol, need_unit)
                return res
            else:
                return None


class Value:
    def __init__(self, val='', unit=None):
        self.derived = False  # 是否是间接计算参数

        try:
            self.value = float(val)  # 将参数的值尽量转换为float类型，以方便计算
        except:
            self.value = val  # 有些值无法转换，因为表示类型的参数可能是字符串
        if unit is None:
            ...
        elif isinstance(unit, str):
            for cls in [Pressure, Temperature, MassFlow, Power, Density, Enthalpy,
                        VolumeFlow, SpecificVolume, AngleUnit, One]:
                if cls().allowed_values is None:
                    logger.debug(f"{cls}的允许单位为空")
                if unit in cls().allowed_values:
                    unit = cls(unit)
                    break
        self.unit: Unit = unit

    def __str__(self):
        return f"{self.value} {self.unit}"

    def get_default_value(self) -> float | str:
        """
        获取默认单位下的值
        """
        try:
            value = float(self.value)
            return value
        except ValueError:
            return self.value  # 也存在一些不是数值型的值，如计算类型等

    def get_value_with_unit(self, unit_str):
        """
        获取指定单位的参数的值

        :param unit_str:
        :return:
        """
        if isinstance(self.unit, Unit):
            try:
                self.value = float(self.value)
            except:
                pass
            return self.unit.transfer_to_unit(self.value, unit_str)
        else:
            return self.value



import socket


def connect_to_server(data):
    client = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    client.connect(('127.0.0.1', 8890))
    data = str(data)
    client.send(data.encode('utf-8'))
    receive_data = client.recv(1024)
    print('recv:', receive_data.decode())
    data = str(data) + 'again'
    client.send(data.encode('utf-8'))


for i in range(10):
    connect_to_server(str(i))



# Sample python script to run an EBSILON model from python
# via the EbsOpen COM API
# Requires pywin32 and a valid EBSILON license with EbsOpen option
# By: Milton Venetos, Wyatt Enterprises, LLC
# Copyright (C) October 18, 2017
# http://www.wyattllc.com
# 注意：引入本文件就需要可用的Ebsilon授权和EbsOpen授权
import os.path
from enum import Enum
from typing import Optional

import win32com.client

from yangke.common.config import logger


class EbsKind(Enum):
    pipe_electric = 1009  # e_uom.epObjectKindPipeElectric
    comp1 = 10001  # e_uom.epObjectKindComp1
    comp2 = 10002  # e_uom.epObjectKindComp2
    comp3 = 10003  # e_uom.epObjectKindComp3
    comp4 = 10004  # e_uom.epObjectKindComp4
    comp5 = 10005  # e_uom.epObjectKindComp5
    comp6 = 10006  # e_uom.epObjectKindComp6
    comp7 = 10007  # e_uom.epObjectKindComp7
    comp8 = 10008  # e_uom.epObjectKindComp8
    comp9 = 10009  # e_uom.epObjectKindComp9
    comp10 = 10010  # e_uom.epObjectKindComp10
    comp11 = 10011  # e_uom.epObjectKindComp11
    comp12 = 10012  # e_uom.epObjectKindComp12
    comp13 = 10013  # e_uom.epObjectKindComp13
    comp14 = 10014  # e_uom.epObjectKindComp14
    comp15 = 10015  # e_uom.epObjectKindComp15
    comp16 = 10016  # e_uom.epObjectKindComp16
    comp17 = 10017  # e_uom.epObjectKindComp17
    comp18 = 10018  # e_uom.epObjectKindComp18
    comp19 = 10019  # e_uom.epObjectKindComp19
    comp20 = 10020  # e_uom.epObjectKindComp20
    comp21 = 10021  # e_uom.epObjectKindComp21
    comp22 = 10022  # e_uom.epObjectKindComp22
    comp23 = 10023  # e_uom.epObjectKindComp23
    comp24 = 10024  # e_uom.epObjectKindComp24
    comp25 = 10025  # e_uom.epObjectKindComp25
    comp26 = 10026  # e_uom.epObjectKindComp26
    comp27 = 10027  # e_uom.epObjectKindComp27
    comp28 = 10028  # e_uom.epObjectKindComp28
    comp29 = 10029  # e_uom.epObjectKindComp29

    comp30 = 10030  # e_uom.epObjectKindComp30
    comp31 = 10031  # e_uom.epObjectKindComp31
    comp32 = 10032  # e_uom.epObjectKindComp32
    comp33 = 10033  # e_uom.epObjectKindComp33
    comp34 = 10034  # e_uom.epObjectKindComp34
    comp35 = 10035  # e_uom.epObjectKindComp35
    comp36 = 10036  # e_uom.epObjectKindComp36
    comp37 = 10037  # e_uom.epObjectKindComp37
    comp38 = 10038  # e_uom.epObjectKindComp38
    comp39 = 10039  # e_uom.epObjectKindComp39

    comp40 = 10040  # e_uom.epObjectKindComp40
    comp41 = 10041  # e_uom.epObjectKindComp41
    comp42 = 10042  # e_uom.epObjectKindComp42
    comp43 = 10043  # e_uom.epObjectKindComp43
    comp44 = 10044  # e_uom.epObjectKindComp44
    comp45 = 10045  # e_uom.epObjectKindComp45
    comp46 = 10046  # e_uom.epObjectKindComp46
    comp47 = 10047  # e_uom.epObjectKindComp47
    comp48 = 10048  # e_uom.epObjectKindComp48
    comp49 = 10049  # e_uom.epObjectKindComp49

    comp50 = 10050  # e_uom.epObjectKindComp50
    comp51 = 10051  # e_uom.epObjectKindComp51
    comp52 = 10052  # e_uom.epObjectKindComp52
    comp53 = 10053  # e_uom.epObjectKindComp53
    comp54 = 10054  # e_uom.epObjectKindComp54
    comp55 = 10055  # e_uom.epObjectKindComp55
    comp56 = 10056  # e_uom.epObjectKindComp56
    comp57 = 10057  # e_uom.epObjectKindComp57
    comp58 = 10058  # e_uom.epObjectKindComp58
    comp59 = 10059  # e_uom.epObjectKindComp59

    comp60 = 10060  # e_uom.epObjectKindComp60
    comp61 = 10061  # e_uom.epObjectKindComp61
    comp62 = 10062  # e_uom.epObjectKindComp62
    comp63 = 10063  # e_uom.epObjectKindComp63
    comp64 = 10064  # e_uom.epObjectKindComp64
    comp65 = 10065  # e_uom.epObjectKindComp65
    comp66 = 10066  # e_uom.epObjectKindComp66
    comp67 = 10067  # e_uom.epObjectKindComp67
    comp68 = 10068  # e_uom.epObjectKindComp68
    comp69 = 10069  # e_uom.epObjectKindComp69

    comp70 = 10070  # e_uom.epObjectKindComp70
    comp71 = 10071  # e_uom.epObjectKindComp71
    comp72 = 10072  # e_uom.epObjectKindComp72
    comp73 = 10073  # e_uom.epObjectKindComp73
    comp74 = 10074  # e_uom.epObjectKindComp74
    comp75 = 10075  # e_uom.epObjectKindComp75
    comp76 = 10076  # e_uom.epObjectKindComp76
    comp77 = 10077  # e_uom.epObjectKindComp77
    comp78 = 10078  # e_uom.epObjectKindComp78
    comp79 = 10079  # e_uom.epObjectKindComp79

    comp80 = 10080  # e_uom.epObjectKindComp80
    comp81 = 10081  # e_uom.epObjectKindComp81
    comp82 = 10082  # e_uom.epObjectKindComp82
    comp83 = 10083  # e_uom.epObjectKindComp83
    comp84 = 10084  # e_uom.epObjectKindComp84
    comp85 = 10085  # e_uom.epObjectKindComp85
    comp86 = 10086  # e_uom.epObjectKindComp86
    comp87 = 10087  # e_uom.epObjectKindComp87
    comp88 = 10088  # e_uom.epObjectKindComp88
    comp89 = 10089  # e_uom.epObjectKindComp89

    comp90 = 10090  # e_uom.epObjectKindComp90
    comp91 = 10091  # e_uom.epObjectKindComp91
    comp92 = 10092  # e_uom.epObjectKindComp92
    comp93 = 10093  # e_uom.epObjectKindComp93
    comp94 = 10094  # e_uom.epObjectKindComp94
    comp95 = 10095  # e_uom.epObjectKindComp95
    comp96 = 10096  # e_uom.epObjectKindComp96
    comp97 = 10097  # e_uom.epObjectKindComp97
    comp98 = 10098  # e_uom.epObjectKindComp98
    comp99 = 10099  # e_uom.epObjectKindComp99

    comp100 = 10100  # e_uom.epObjectKindComp100
    comp101 = 10101  # e_uom.epObjectKindComp101
    comp102 = 10102  # e_uom.epObjectKindComp102
    comp103 = 10103  # e_uom.epObjectKindComp103
    comp104 = 10104  # e_uom.epObjectKindComp104
    comp105 = 10105  # e_uom.epObjectKindComp105
    comp106 = 10106  # e_uom.epObjectKindComp106
    comp107 = 10107  # e_uom.epObjectKindComp107
    comp108 = 10108  # e_uom.epObjectKindComp108
    comp109 = 10109  # e_uom.epObjectKindComp109

    comp110 = 10110  # e_uom.epObjectKindComp110
    comp111 = 10111  # e_uom.epObjectKindComp111
    comp112 = 10112  # e_uom.epObjectKindComp112
    comp113 = 10113  # e_uom.epObjectKindComp113
    comp114 = 10114  # e_uom.epObjectKindComp114
    comp115 = 10115  # e_uom.epObjectKindComp115
    comp116 = 10116  # e_uom.epObjectKindComp116
    comp117 = 10117  # e_uom.epObjectKindComp117
    comp118 = 10118  # e_uom.epObjectKindComp118
    comp119 = 10119  # e_uom.epObjectKindComp119

    comp120 = 10120  # e_uom.epObjectKindComp120
    comp121 = 10121  # e_uom.epObjectKindComp121
    comp122 = 10122  # e_uom.epObjectKindComp122
    comp123 = 10123  # e_uom.epObjectKindComp123
    comp124 = 10124  # e_uom.epObjectKindComp124
    comp125 = 10125  # e_uom.epObjectKindComp125
    comp126 = 10126  # e_uom.epObjectKindComp126
    comp127 = 10127  # e_uom.epObjectKindComp127
    comp128 = 10128  # e_uom.epObjectKindComp128
    comp129 = 10129  # e_uom.epObjectKindComp129

    comp130 = 10130  # e_uom.epObjectKindComp130
    comp131 = 10131  # e_uom.epObjectKindComp131
    comp132 = 10132  # e_uom.epObjectKindComp132
    comp133 = 10133  # e_uom.epObjectKindComp133
    comp134 = 10134  # e_uom.epObjectKindComp134
    comp135 = 10135  # e_uom.epObjectKindComp135
    comp136 = 10136  # e_uom.epObjectKindComp136
    comp137 = 10137  # e_uom.epObjectKindComp137
    comp138 = 10138  # e_uom.epObjectKindComp138
    try:  # 12版ebsilon不存在以下组件
        comp139 = 10139  # e_uom.epObjectKindComp139
        comp140 = 10140  # e_uom.epObjectKindComp140
        comp141 = 10141  # e_uom.epObjectKindComp141
        comp142 = 10142  # e_uom.epObjectKindComp142
        comp143 = 10143  # e_uom.epObjectKindComp143
        comp144 = 10144  # e_uom.epObjectKindComp144
        comp145 = 10145  # e_uom.epObjectKindComp145
        comp146 = 10146  # e_uom.epObjectKindComp146
        comp147 = 10147  # e_uom.epObjectKindComp147
        comp148 = 10148  # e_uom.epObjectKindComp148
        comp149 = 10149  # e_uom.epObjectKindComp149

        comp150 = 10150  # e_uom.epObjectKindComp150
        comp151 = 10151  # e_uom.epObjectKindComp151
        comp152 = 10152  # e_uom.epObjectKindComp152
        comp153 = 10153  # e_uom.epObjectKindComp153
        comp154 = 10154  # e_uom.epObjectKindComp154
        comp155 = 10155  # e_uom.epObjectKindComp155
        comp156 = 10156  # e_uom.epObjectKindComp156
        comp157 = 10157  # e_uom.epObjectKindComp157
    except AttributeError:
        pass


class EbsUnits(Enum):
    """
    Ebsilon的单位系统，包含了Ebsilon软件中定义的所有单位常量，在读写组件参数时会用到。参见ebsilon帮助文件EpUnit Enumeration
    """
    global e_uom
    # 流量
    lb_h: "lb_h" = 326  # e_uom.epUNIT_lb_h  # 磅/小时，流量单位
    kg_s: "kg/s" = 6  # e_uom.epUNIT_kg_s  # 流量单位kg/s
    # 功率
    MW: "MW" = 100  # e_uom.epUNIT_MW  # 功率单位，MW
    kW = 7  # e_uom.epUNIT_kW
    W = 216  # e_uom.epUNIT_W
    # 压力
    bar = 3  # e_uom.epUNIT_bar
    Pa = 203  # e_uom.epUNIT_Pa
    kPa = 217  # e_uom.epUNIT_kPa
    MPa = 123  # e_uom.epUNIT_MPa
    # 温度
    C = 372  # e_uom.epUNIT_C
    K = 12  # e_uom.epUNIT_K
    # 能量
    J = 297  # e_uom.epUNIT_J
    kJ = 296  # e_uom.epUNIT_kJ
    MJ = 298  # e_uom.epUNIT_MJ

    # 密度
    g_m3 = 301  # e_uom.epUNIT_g_m3
    # 转速
    unit_1_min = 17  # 试验测试得到的值
    # 其他
    int = 321  # e_uom.epUNIT_INTEGRAL
    text = 53  # e_uom.epUNIT_TEXT
    access_actual = -10  # e_uom.epUNITACCESS_ACTUAL
    access_si = -11  # e_uom.epUNITACCESS_SI
    ACCESS_Imperial = -12  # e_uom.epUNITACCESS_Imperial
    ACCESS_USC = -13  # e_uom.epUNITACCESS_USC
    # ERROR = e_uom.epUNIT_ERROR
    INVALID = 0  # e_uom.epUNIT_INVALID
    NONE = 1  # e_uom.epUNIT_NONE
    unit_1 = 2  # e_uom.epUNIT_1
    GrdC = 4  # e_uom.epUNIT_GrdC
    kJ_kg = 5  # e_uom.epUNIT_kJ_kg
    m3_kg = 8  # e_uom.epUNIT_m3_kg
    m3_s = 9  # e_uom.epUNIT_m3_s
    kmol_kmol = 13  # e_uom.epUNIT_kmol_kmol
    kg_kg = 14  # e_uom.epUNIT_kg_kg
    kW_K = 15  # e_uom.epUNIT_kW_K
    W_m2K = 16  # e_uom.epUNIT_W_m2K
    kJ_kWh = 18  # e_uom.epUNIT_kJ_kWh
    kJ_m3 = 21  # e_uom.epUNIT_kJ_m3
    kJ_m3K = 22  # e_uom.epUNIT_kJ_m3K
    kg_m3 = 23  # e_uom.epUNIT_kg_m3
    m = 24  # e_uom.epUNIT_m
    kJ_kgK_cp = 26  # e_uom.epUNIT_kJ_kgK_cp
    m2 = 27  # e_uom.epUNIT_m2
    kJ_kgK = 28  # e_uom.epUNIT_kJ_kgK
    kg_kg_x = 29  # e_uom.epUNIT_kg_kg_x
    kg_kg_xg = 30  # e_uom.epUNIT_kg_kg_xg
    kg_kmol = 31  # e_uom.epUNIT_kg_kmol
    kJ_kg_ncv = 32  # e_uom.epUNIT_kJ_kg_ncv
    m_s = 33  # e_uom.epUNIT_m_s
    kg_kg_x_rg = 34  # e_uom.epUNIT_kg_kg_x_rg
    FTYP_8 = 35  # e_uom.epUNIT_FTYP_8
    FTYP_9 = 36  # e_uom.epUNIT_FTYP_9
    mg_Nm3 = 37  # e_uom.epUNIT_mg_Nm3
    EUR_h = 38  # e_uom.epUNIT_EUR_h
    kW_kg = 39  # e_uom.epUNIT_kW_kg
    _1_m6 = 40  # e_uom.epUNIT_1_m6
    A = 41  # e_uom.epUNIT_A
    EUR_kWh = 42  # e_uom.epUNIT_EUR_kWh
    EUR_kg = 43  # e_uom.epUNIT_EUR_kg
    V = 44  # e_uom.epUNIT_V
    m3_m3 = 45  # e_uom.epUNIT_m3_m3
    kg = 46  # e_uom.epUNIT_kg
    EUR = 47  # e_uom.epUNIT_EUR
    m3 = 48  # e_uom.epUNIT_m3
    ph = 49  # e_uom.epUNIT_ph
    m2K_W = 51  # e_uom.epUNIT_m2K_W
    W_m2 = 52  # e_uom.epUNIT_W_m2
    TEXT = 53  # e_uom.epUNIT_TEXT
    Grd = 54  # e_uom.epUNIT_Grd
    m_geopot = 59  # e_uom.epUNIT_m_geopot
    t_h = 101  # e_uom.epUNIT_t_h
    GrdF = 102  # e_uom.epUNIT_GrdF
    Prz = 103  # e_uom.epUNIT_Prz
    psia = 104  # e_uom.epUNIT_psia
    btu_lb = 105  # e_uom.epUNIT_btu_lb
    klb_h = 106  # e_uom.epUNIT_klb_h
    ft3_lb = 107  # e_uom.epUNIT_ft3_lb
    Mft3_h = 108  # e_uom.epUNIT_Mft3_h
    R = 109  # e_uom.epUNIT_R
    lb_lb = 110  # e_uom.epUNIT_lb_lb
    kbtu_hF = 111  # e_uom.epUNIT_kbtu_hF
    btu_ft2hF = 112  # e_uom.epUNIT_btu_ft2hF
    lb_ft3 = 113  # e_uom.epUNIT_lb_ft3
    btu_lbF = 114  # e_uom.epUNIT_btu_lbF
    btu_kWh = 115  # e_uom.epUNIT_btu_kWh
    btu_ft3 = 116  # e_uom.epUNIT_btu_ft3
    btu_ft3F = 117  # e_uom.epUNIT_btu_ft3F
    ft = 118  # e_uom.epUNIT_ft
    rpm = 119  # e_uom.epUNIT_rpm
    hp = 120  # e_uom.epUNIT_hp
    ft2 = 121  # e_uom.epUNIT_ft2
    m3_h = 122  # e_uom.epUNIT_m3_h
    btu_lbF_cp = 124  # e_uom.epUNIT_btu_lbF_cp
    mbar = 125  # e_uom.epUNIT_mbar
    lb_lb_x = 126  # e_uom.epUNIT_lb_lb_x
    lb_lb_xg = 127  # e_uom.epUNIT_lb_lb_xg
    lb_kmol = 128  # e_uom.epUNIT_lb_kmol
    btu_lb_ncv = 129  # e_uom.epUNIT_btu_lb_ncv
    ft_s = 130  # e_uom.epUNIT_ft_s
    lb_lb_x_rg = 131  # e_uom.epUNIT_lb_lb_x_rg
    CENT_min = 132  # e_uom.epUNIT_CENT_min
    DM_h = 133  # e_uom.epUNIT_DM_h
    PFENNIG_min = 134  # e_uom.epUNIT_PFENNIG_min
    Prz_x = 137  # e_uom.epUNIT_Prz_x
    Prz_xg = 138  # e_uom.epUNIT_Prz_xg
    g_mol = 139  # e_uom.epUNIT_g_mol
    ppm = 140  # e_uom.epUNIT_ppm
    ppm_xg = 141  # e_uom.epUNIT_ppm_xg
    MW_kg = 142  # e_uom.epUNIT_MW_kg
    hp_lb = 143  # e_uom.epUNIT_hp_lb
    mg_m3 = 144  # e_uom.epUNIT_mg_m3
    _1_ft6 = 145  # e_uom.epUNIT_1_ft6
    kg_h = 146  # e_uom.epUNIT_kg_h
    mA = 147  # e_uom.epUNIT_mA
    t_s = 148  # e_uom.epUNIT_t_s
    EUR_MWh = 149  # e_uom.epUNIT_EUR_MWh
    CENT_kWh = 150  # e_uom.epUNIT_CENT_kWh
    EUR_t = 151  # e_uom.epUNIT_EUR_t
    CENT_kg = 152  # e_uom.epUNIT_CENT_kg
    kV = 153  # e_uom.epUNIT_kV
    g_kg = 158  # e_uom.epUNIT_g_kg
    mg_kg = 159  # e_uom.epUNIT_mg_kg
    Prz_vol = 160  # e_uom.epUNIT_Prz_vol
    ppm_vol = 161  # e_uom.epUNIT_ppm_vol
    ft3_ft3 = 162  # e_uom.epUNIT_ft3_ft3
    t = 164  # e_uom.epUNIT_t
    g = 165  # e_uom.epUNIT_g
    mg = 166  # e_uom.epUNIT_mg
    lb = 167  # e_uom.epUNIT_lb
    klb = 168  # e_uom.epUNIT_klb
    CENT = 169  # e_uom.epUNIT_CENT
    DM = 170  # e_uom.epUNIT_DM
    Pfennig = 171  # e_uom.epUNIT_Pfennig
    l_min = 172  # e_uom.epUNIT_l_min
    l_h = 173  # e_uom.epUNIT_l_h
    mV = 174  # e_uom.epUNIT_mV
    mm = 176  # e_uom.epUNIT_mm
    cm = 177  # e_uom.epUNIT_cm
    km = 178  # e_uom.epUNIT_km
    yd = 179  # e_uom.epUNIT_yd
    inch = 180  # e_uom.epUNIT_in
    mi = 181  # e_uom.epUNIT_mi
    mm2 = 182  # e_uom.epUNIT_mm2
    cm2 = 183  # e_uom.epUNIT_cm2
    L = 184  # e_uom.epUNIT_l
    km2 = 185  # e_uom.epUNIT_km2
    yd2 = 186  # e_uom.epUNIT_yd2
    in2 = 187  # e_uom.epUNIT_in2
    mi2 = 188  # e_uom.epUNIT_mi2
    ft3 = 189  # e_uom.epUNIT_ft3
    mm3 = 190  # e_uom.epUNIT_mm3
    cm3 = 191  # e_uom.epUNIT_cm3
    km3 = 192  # e_uom.epUNIT_km3
    yd3 = 193  # e_uom.epUNIT_yd3
    in3 = 194  # e_uom.epUNIT_in3
    mi3 = 195  # e_uom.epUNIT_mi3
    gal = 196  # e_uom.epUNIT_gal
    kg_Nm3 = 197  # e_uom.epUNIT_kg_Nm3
    tm3_h = 198  # e_uom.epUNIT_tm3_h
    Mm3_h = 199  # e_uom.epUNIT_Mm3_h
    mmWS = 201  # e_uom.epUNIT_mmWS
    mWS = 202  # e_uom.epUNIT_mWS
    kA = 204  # e_uom.epUNIT_kA
    l_s = 205  # e_uom.epUNIT_l_s
    Hz = 206  # e_uom.epUNIT_Hz
    mmHg = 207  # e_uom.epUNIT_mmHg
    mg_l = 208  # e_uom.epUNIT_mg_l
    at = 209  # e_uom.epUNIT_at
    kcal_kg = 210  # e_uom.epUNIT_kcal_kg
    kcal_kg_ncv = 211  # e_uom.epUNIT_kcal_kg_ncv
    kcal_kgK = 212  # e_uom.epUNIT_kcal_kgK
    kcal_kgK_cp = 213  # e_uom.epUNIT_kcal_kgK_cp
    kW_m2 = 214  # e_uom.epUNIT_kW_m2
    kcal_m2h = 215  # e_uom.epUNIT_kcal_m2h
    Nm_kg = 218  # e_uom.epUNIT_Nm_kg
    kVA = 219  # e_uom.epUNIT_kVA
    MVA = 220  # e_uom.epUNIT_MVA
    VA = 221  # e_uom.epUNIT_VA
    kVAr = 222  # e_uom.epUNIT_kVAr
    MVAr = 223  # e_uom.epUNIT_MVAr
    VAr = 224  # e_uom.epUNIT_VAr
    atm = 225  # e_uom.epUNIT_atm
    g_s = 226  # e_uom.epUNIT_g_s
    kg_ms = 227  # e_uom.epUNIT_kg_ms
    W_mK = 228  # e_uom.epUNIT_W_mK
    inHg = 229  # e_uom.epUNIT_inHg
    ft_geopot = 230  # e_uom.epUNIT_ft_geopot
    km_geopot = 231  # e_uom.epUNIT_km_geopot
    yd_geopot = 232  # e_uom.epUNIT_yd_geopot
    mi_geopot = 233  # e_uom.epUNIT_mi_geopot
    rad = 234  # e_uom.epUNIT_rad
    _1_Grd = 235  # e_uom.epUNIT_1_Grd
    _1_Grd2 = 236  # e_uom.epUNIT_1_Grd2
    _1_Grd3 = 237  # e_uom.epUNIT_1_Grd3
    _1_Grd4 = 238  # e_uom.epUNIT_1_Grd4
    _1_Grd5 = 239  # e_uom.epUNIT_1_Grd5
    _1_rad = 240  # e_uom.epUNIT_1_rad
    _1_rad2 = 241  # e_uom.epUNIT_1_rad2
    _1_rad3 = 242  # e_uom.epUNIT_1_rad3
    _1_rad4 = 243  # e_uom.epUNIT_1_rad4
    _1_rad5 = 244  # e_uom.epUNIT_1_rad5
    _1_K = 245  # e_uom.epUNIT_1_K
    _1_K2 = 246  # e_uom.epUNIT_1_K2
    _1_K3 = 247  # e_uom.epUNIT_1_K3
    _1_K4 = 248  # e_uom.epUNIT_1_K4
    _1_R = 249  # e_uom.epUNIT_1_R
    _1_R2 = 250  # e_uom.epUNIT_1_R2
    _1_R3 = 251  # e_uom.epUNIT_1_R3
    _1_R4 = 252  # e_uom.epUNIT_1_R4
    W_m = 253  # e_uom.epUNIT_W_m
    kW_m = 254  # e_uom.epUNIT_kW_m
    kcal_mh = 255  # e_uom.epUNIT_kcal_mh
    GrdK = 256  # e_uom.epUNIT_GrdK
    s = 257  # e_uom.epUNIT_s
    min = 258  # e_uom.epUNIT_min
    h = 259  # e_uom.epUNIT_h
    d = 260  # e_uom.epUNIT_d
    K_m = 261  # e_uom.epUNIT_K_m
    kJ_kgm = 262  # e_uom.epUNIT_kJ_kgm
    K_cm = 263  # e_uom.epUNIT_K_cm
    K_mm = 264  # e_uom.epUNIT_K_mm
    K_km = 265  # e_uom.epUNIT_K_km
    K_ft = 266  # e_uom.epUNIT_K_ft
    K_yd = 267  # e_uom.epUNIT_K_yd
    R_ft = 268  # e_uom.epUNIT_R_ft
    R_yd = 269  # e_uom.epUNIT_R_yd
    btu_lbft = 270  # e_uom.epUNIT_btu_lbft
    datetime = 271  # e_uom.epUNIT_datetime
    kW_kgK = 272  # e_uom.epUNIT_kW_kgK
    W_kgK = 273  # e_uom.epUNIT_W_kgK
    W_gK = 274  # e_uom.epUNIT_W_gK
    kbtu_lbhF = 275  # e_uom.epUNIT_kbtu_lbhF
    bar_m = 276  # e_uom.epUNIT_bar_m
    mbar_m = 277  # e_uom.epUNIT_mbar_m
    mbar_cm = 278  # e_uom.epUNIT_mbar_cm
    psia_ft = 279  # e_uom.epUNIT_psia_ft
    kt = 280  # e_uom.epUNIT_kt
    Mt = 281  # e_uom.epUNIT_Mt
    mN_m = 282  # e_uom.epUNIT_mN_m
    N_m = 283  # e_uom.epUNIT_N_m
    W_mK2 = 284  # e_uom.epUNIT_W_mK2
    W_mK3 = 285  # e_uom.epUNIT_W_mK3
    W_mK4 = 286  # e_uom.epUNIT_W_mK4
    m_K = 287  # e_uom.epUNIT_m_K
    m_K2 = 288  # e_uom.epUNIT_m_K2
    m2_s = 289  # e_uom.epUNIT_m2_s
    ft2_s = 290  # e_uom.epUNIT_ft2_s
    mm2_s = 291  # e_uom.epUNIT_mm2_s
    cm2_s = 292  # e_uom.epUNIT_cm2_s
    in2_s = 293  # e_uom.epUNIT_in2_s
    Pas = 294  # e_uom.epUNIT_Pas
    mPas = 295  # e_uom.epUNIT_mPas
    kWh = 299  # e_uom.epUNIT_kWh
    MWh = 300  # e_uom.epUNIT_MWh
    Nm3_s = 302  # e_uom.epUNIT_Nm3_s
    Nm3_h = 303  # e_uom.epUNIT_Nm3_h
    kNm3_h = 304  # e_uom.epUNIT_kNm3_h
    kg_m3K = 305  # e_uom.epUNIT_kg_m3K
    lb_ft3R = 306  # e_uom.epUNIT_lb_ft3R
    g_m3K = 307  # e_uom.epUNIT_g_m3K
    kg_lK = 308  # e_uom.epUNIT_kg_lK
    g_lK = 309  # e_uom.epUNIT_g_lK
    g_cm3K = 310  # e_uom.epUNIT_g_cm3K
    kJ_kgK2 = 311  # e_uom.epUNIT_kJ_kgK2
    btu_lbR2 = 312  # e_uom.epUNIT_btu_lbR2
    kg_l = 313  # e_uom.epUNIT_kg_l
    g_l = 314  # e_uom.epUNIT_g_l
    g_cm3 = 315  # e_uom.epUNIT_g_cm3
    kcal_kWh = 316  # e_uom.epUNIT_kcal_kWh
    mrad = 317  # e_uom.epUNIT_mrad
    kg2_kJs = 318  # e_uom.epUNIT_kg2_kJs
    lb2_btus = 319  # e_uom.epUNIT_lb2_btus
    kW_m2K = 320  # e_uom.epUNIT_kW_m2K
    INTEGRAL = 321  # e_uom.epUNIT_INTEGRAL
    P = 322  # e_uom.epUNIT_P
    cP = 323  # e_uom.epUNIT_cP
    St = 324  # e_uom.epUNIT_St
    cSt = 325  # e_uom.epUNIT_cSt
    lb_s = 327  # e_uom.epUNIT_lb_s
    klb_s = 328  # e_uom.epUNIT_klb_s
    oz_s = 329  # e_uom.epUNIT_oz_s
    oz_h = 330  # e_uom.epUNIT_oz_h
    y3_s = 331  # e_uom.epUNIT_y3_s
    ft3_s = 332  # e_uom.epUNIT_ft3_s
    in3_s = 333  # e_uom.epUNIT_in3_s
    gal_s = 334  # e_uom.epUNIT_gal_s
    y3_h = 336  # e_uom.epUNIT_y3_h
    ft3_h = 337  # e_uom.epUNIT_ft3_h
    in3_h = 338  # e_uom.epUNIT_in3_h
    gal_h = 339  # e_uom.epUNIT_gal_h
    kft3_s = 340  # e_uom.epUNIT_kft3_s
    kft3_h = 341  # e_uom.epUNIT_kft3_h
    m_h = 342  # e_uom.epUNIT_m_h
    km_h = 343  # e_uom.epUNIT_km_h
    y_s = 344  # e_uom.epUNIT_y_s
    in_s = 345  # e_uom.epUNIT_in_s
    y_h = 346  # e_uom.epUNIT_y_h
    ft_h = 347  # e_uom.epUNIT_ft_h
    mi_h = 348  # e_uom.epUNIT_mi_h
    W_mGrdC = 349  # e_uom.epUNIT_W_mGrdC
    W_mGrdC2 = 350  # e_uom.epUNIT_W_mGrdC2
    W_mGrdC3 = 351  # e_uom.epUNIT_W_mGrdC3
    W_mGrdC4 = 352  # e_uom.epUNIT_W_mGrdC4
    m_GrdC = 353  # e_uom.epUNIT_m_GrdC
    m_GrdC2 = 354  # e_uom.epUNIT_m_GrdC2
    bars_kg = 356  # e_uom.epUNIT_bars_kg
    barkg_kJ = 357  # e_uom.epUNIT_barkg_kJ
    barK_kW = 358  # e_uom.epUNIT_barK_kW
    ppm_kg = 359  # e_uom.epUNIT_ppm_kg
    ppm_x = 360  # e_uom.epUNIT_ppm_x
    ppm_x_rg = 361  # e_uom.epUNIT_ppm_x_rg
    Prz_kg = 362  # e_uom.epUNIT_Prz_kg
    Prz_x_rg = 363  # e_uom.epUNIT_Prz_x_rg
    g_kg_x = 364  # e_uom.epUNIT_g_kg_x
    mg_kg_x = 365  # e_uom.epUNIT_mg_kg_x
    g_kg_xg = 366  # e_uom.epUNIT_g_kg_xg
    mg_kg_xg = 367  # e_uom.epUNIT_mg_kg_xg
    g_kg_x_rg = 368  # e_uom.epUNIT_g_kg_x_rg
    mg_kg_x_rg = 369  # e_uom.epUNIT_mg_kg_x_rg
    _1_s = 370  # e_uom.epUNIT_1_s
    _1_h = 371  # e_uom.epUNIT_1_h
    F = 373  # e_uom.epUNIT_F
    inWS = 374  # e_uom.epUNIT_inWS
    cmWS = 375  # e_uom.epUNIT_cmWS
    kcal = 376  # e_uom.epUNIT_kcal
    cal = 377  # e_uom.epUNIT_cal
    kcal_h = 378  # e_uom.epUNIT_kcal_h
    lb_lbmole = 379  # e_uom.epUNIT_lb_lbmole
    lbmole_lbmole = 380  # e_uom.epUNIT_lbmole_lbmole
    kmol_lbmole = 381  # e_uom.epUNIT_kmol_lbmole
    lbmole_kmol = 382  # e_uom.epUNIT_lbmole_kmol
    MJ_m3 = 383  # e_uom.epUNIT_MJ_m3
    btu = 384  # e_uom.epUNIT_btu
    N = 385  # e_uom.epUNIT_N
    kN = 386  # e_uom.epUNIT_kN
    mN = 387  # e_uom.epUNIT_mN
    lbf = 388  # e_uom.epUNIT_lbf
    kp = 389  # e_uom.epUNIT_kp
    PS = 390  # e_uom.epUNIT_PS
    Torr = 391  # e_uom.epUNIT_Torr
    unit_1_m = 392  # e_uom.epUNIT_1_m
    unit_1_ft = 393  # e_uom.epUNIT_1_ft
    unit_1_mm = 394  # e_uom.epUNIT_1_mm
    unit_1_cm = 395  # e_uom.epUNIT_1_cm
    unit_1_km = 396  # e_uom.epUNIT_1_km
    unit_1_yd = 397  # e_uom.epUNIT_1_yd
    unit_1_in = 398  # e_uom.epUNIT_1_in
    unit_1_mi = 399  # e_uom.epUNIT_1_mi
    btu_h = 400  # e_uom.epUNIT_btu_h
    kbtu_h = 401  # e_uom.epUNIT_kbtu_h
    Mbtu_h = 402  # e_uom.epUNIT_Mbtu_h
    m2_W = 403  # e_uom.epUNIT_m2_W
    kJ_Nm3 = 404  # e_uom.epUNIT_kJ_Nm3
    MJ_Nm3 = 405  # e_uom.epUNIT_MJ_Nm3
    MJ_SCM = 406  # e_uom.epUNIT_MJ_SCM
    btu_SCF = 407  # e_uom.epUNIT_btu_SCF
    oz = 408  # e_uom.epUNIT_oz
    gr = 409  # e_uom.epUNIT_gr
    mg_SCM = 410  # e_uom.epUNIT_mg_SCM
    gr_SCF = 411  # e_uom.epUNIT_gr_SCF
    SCM_s = 412  # e_uom.epUNIT_SCM_s
    SCF_s = 413  # e_uom.epUNIT_SCF_s
    PATH = 414  # e_uom.epUNIT_PATH
    FOLDER = 415  # e_uom.epUNIT_FOLDER
    KERNELEXPRESSION = 416  # e_uom.epUNIT_KERNELEXPRESSION
    unit_1_kg = 417  # e_uom.epUNIT_1_kg
    kg_kg_1xg = 418  # e_uom.epUNIT_kg_kg_1xg
    mol_mol = 419  # e_uom.epUNIT_mol_mol
    mol_prz = 420  # e_uom.epUNIT_mol_prz
    mass_prz = 421  # e_uom.epUNIT_mass_prz
    mol_ppm = 422  # e_uom.epUNIT_mol_ppm
    ftH2O = 423  # e_uom.epUNIT_ftH2O
    mTorr = 424  # e_uom.epUNIT_mTorr
    oz_in2 = 425  # e_uom.epUNIT_oz_in2
    hPa = 426  # e_uom.epUNIT_hPa
    cmHg = 427  # e_uom.epUNIT_cmHg
    kg_cm2 = 428  # e_uom.epUNIT_kg_cm2
    g_cm2 = 429  # e_uom.epUNIT_g_cm2
    shton_s = 430  # e_uom.epUNIT_shton_s
    shton_h = 431  # e_uom.epUNIT_shton_h
    lgton_s = 432  # e_uom.epUNIT_lgton_s
    lgton_h = 433  # e_uom.epUNIT_lgton_h
    DOLLAR_h = 434  # e_uom.epUNIT_DOLLAR_h
    POUND_h = 435  # e_uom.epUNIT_POUND_h
    YEN_h = 436  # e_uom.epUNIT_YEN_h
    DOLLAR_kWh = 437  # e_uom.epUNIT_DOLLAR_kWh
    POUND_kWh = 438  # e_uom.epUNIT_POUND_kWh
    YEN_kWh = 439  # e_uom.epUNIT_YEN_kWh
    DOLLAR_kg = 440  # e_uom.epUNIT_DOLLAR_kg
    POUND_kg = 441  # e_uom.epUNIT_POUND_kg
    YEN_kg = 442  # e_uom.epUNIT_YEN_kg
    DOLLAR = 443  # e_uom.epUNIT_DOLLAR
    POUND = 444  # e_uom.epUNIT_POUND
    YEN = 445  # e_uom.epUNIT_YEN
    DOLLAR_s = 446  # e_uom.epUNIT_DOLLAR_s
    POUND_s = 447  # e_uom.epUNIT_POUND_s
    YEN_s = 448  # e_uom.epUNIT_YEN_s
    EUR_s = 449  # e_uom.epUNIT_EUR_s
    CENT_s = 450  # e_uom.epUNIT_CENT_s
    CENT_h = 451  # e_uom.epUNIT_CENT_h
    kNm3_s = 452  # e_uom.epUNIT_kNm3_s
    kJ_mol = 453  # e_uom.epUNIT_kJ_mol
    J_mol = 454  # e_uom.epUNIT_J_mol
    btu_mol = 455  # e_uom.epUNIT_btu_mol
    kcal_mol = 456  # e_uom.epUNIT_kcal_mol
    gal_min = 457  # e_uom.epUNIT_gal_min
    mSQRT_K_W = 458  # e_uom.epUNIT_mSQRT_K_W
    mSQRT_K_kW = 459  # e_uom.epUNIT_mSQRT_K_kW
    ftSQRT_Rk_W = 460  # e_uom.epUNIT_ftSQRT_Rk_W
    impgal = 461  # e_uom.epUNIT_impgal
    impgal_s = 462  # e_uom.epUNIT_impgal_s
    impgal_min = 463  # e_uom.epUNIT_impgal_min
    impgal_h = 464  # e_uom.epUNIT_impgal_h
    impgal_d = 465  # e_uom.epUNIT_impgal_d
    Mimpgal_d = 466  # e_uom.epUNIT_Mimpgal_d
    A_K = 467  # e_uom.epUNIT_A_K
    mA_K = 468  # e_uom.epUNIT_mA_K
    V_K = 469  # e_uom.epUNIT_V_K
    mV_K = 470  # e_uom.epUNIT_mV_K
    A_F = 471  # e_uom.epUNIT_A_F
    mA_F = 472  # e_uom.epUNIT_mA_F
    V_F = 473  # e_uom.epUNIT_V_F
    mV_F = 474  # e_uom.epUNIT_mV_F
    Ohm = 475  # e_uom.epUNIT_Ohm
    kOhm = 476  # e_uom.epUNIT_kOhm
    MOhm = 477  # e_uom.epUNIT_MOhm
    Farad = 478  # e_uom.epUNIT_Farad
    mFarad = 479  # e_uom.epUNIT_mFarad
    muFarad = 480  # e_uom.epUNIT_muFarad
    nFarad = 481  # e_uom.epUNIT_nFarad
    pFarad = 482  # e_uom.epUNIT_pFarad
    Henry = 483  # e_uom.epUNIT_Henry
    mHenry = 484  # e_uom.epUNIT_mHenry
    kHenry = 485  # e_uom.epUNIT_kHenry
    muHenry = 486  # e_uom.epUNIT_muHenry
    kJ_kgs = 487  # e_uom.epUNIT_kJ_kgs
    kJ_kgh = 488  # e_uom.epUNIT_kJ_kgh
    kWs_kg = 489  # e_uom.epUNIT_kWs_kg
    kWh_kg = 490  # e_uom.epUNIT_kWh_kg
    kWs_m3 = 491  # e_uom.epUNIT_kWs_m3
    kWh_m3 = 492  # e_uom.epUNIT_kWh_m3
    GrdR = 493  # e_uom.epUNIT_GrdR
    try:  # 12版ebsilon不存在以下单位
        Nm = 494  # e_uom.epUNIT_Nm
        kNm = 495  # e_uom.epUNIT_kNm
        lbf_ft = 496  # e_uom.epUNIT_lbf_ft
        ozf_in = 497  # e_uom.epUNIT_ozf_in
        lbf_in = 498  # e_uom.epUNIT_lbf_in
        ozf_ft = 499  # e_uom.epUNIT_ozf_ft
        dynm = 500  # e_uom.epUNIT_dynm
        Wh = 501  # e_uom.epUNIT_Wh
        Wh_kg = 502  # e_uom.epUNIT_Wh_kg
        Wh_m3 = 503  # e_uom.epUNIT_Wh_m3
        kJ_m2 = 504  # e_uom.epUNIT_kJ_m2
        kWh_m2 = 505  # e_uom.epUNIT_kWh_m2
        Wh_m2 = 506  # e_uom.epUNIT_Wh_m2
        btu_ft2 = 507  # e_uom.epUNIT_btu_ft2
        W_m3 = 508  # e_uom.epUNIT_W_m3
        kW_m3 = 509  # e_uom.epUNIT_kW_m3
        m2K_kW = 510  # e_uom.epUNIT_m2K_kW
        ft2hF_btu = 511  # e_uom.epUNIT_ft2hF_btu
        MJ_kg = 512  # e_uom.epUNIT_MJ_kg
        GJ_kg = 513  # e_uom.epUNIT_GJ_kg
        J_kg = 514  # e_uom.epUNIT_J_kg
        MJ_kWh = 515  # e_uom.epUNIT_MJ_kWh
        GJ_kWh = 516  # e_uom.epUNIT_GJ_kWh
        J_kWh = 517  # e_uom.epUNIT_J_kWh
        GJ_m3 = 518  # e_uom.epUNIT_GJ_m3
        J_m3 = 519  # e_uom.epUNIT_J_m3
        MJ_m3K = 520  # e_uom.epUNIT_MJ_m3K
        GJ_m3K = 521  # e_uom.epUNIT_GJ_m3K
        J_m3K = 522  # e_uom.epUNIT_J_m3K
        MJ_kgK_cp = 523  # e_uom.epUNIT_MJ_kgK_cp
        GJ_kgK_cp = 524  # e_uom.epUNIT_GJ_kgK_cp
        J_kgK_cp = 525  # e_uom.epUNIT_J_kgK_cp
        MJ_kgK = 526  # e_uom.epUNIT_MJ_kgK
        GJ_kgK = 527  # e_uom.epUNIT_GJ_kgK
        J_kgK = 528  # e_uom.epUNIT_J_kgK
        MJ_kg_ncv = 529  # e_uom.epUNIT_MJ_kg_ncv
        GJ_kg_ncv = 530  # e_uom.epUNIT_GJ_kg_ncv
        J_kg_ncv = 531  # e_uom.epUNIT_J_kg_ncv
        MJ_kgm = 532  # e_uom.epUNIT_MJ_kgm
        GJ_kgm = 533  # e_uom.epUNIT_GJ_kgm
        J_kgm = 534  # e_uom.epUNIT_J_kgm
        GW = 536  # e_uom.epUNIT_GW
        TW = 537  # e_uom.epUNIT_TW
        GJ_h = 538  # e_uom.epUNIT_GJ_h
        MJ_h = 539  # e_uom.epUNIT_MJ_h
        kJ_h = 540  # e_uom.epUNIT_kJ_h
        J_h = 541  # e_uom.epUNIT_J_h
        J_s = 542  # e_uom.epUNIT_J_s
        mg_s = 543  # e_uom.epUNIT_mg_s
        g_h = 544  # e_uom.epUNIT_g_h
        mg_h = 545  # e_uom.epUNIT_mg_h
        mug_s = 546  # e_uom.epUNIT_mug_s
        mug_h = 547  # e_uom.epUNIT_mug_h
        mug = 548  # e_uom.epUNIT_mug
        mum = 549  # e_uom.epUNIT_mum
        ml = 550  # e_uom.epUNIT_ml
        ml_s = 551  # e_uom.epUNIT_ml_s
        cm3_s = 552  # e_uom.epUNIT_cm3_s
        mm3_s = 553  # e_uom.epUNIT_mm3_s
        kW_mK = 554  # e_uom.epUNIT_kW_mK
        l_kg = 555  # e_uom.epUNIT_l_kg
        btu_hftK = 556  # e_uom.epUNIT_btu_hftK
        W_mm = 557  # e_uom.epUNIT_W_mm
        W_cm = 558  # e_uom.epUNIT_W_cm
        btu_hft = 559  # e_uom.epUNIT_btu_hft
        W_ft = 560  # e_uom.epUNIT_W_ft
        W_cm2 = 561  # e_uom.epUNIT_W_cm2
        btu_hft2 = 562  # e_uom.epUNIT_btu_hft2
        W_ft2 = 563  # e_uom.epUNIT_W_ft2
        MW_m3 = 564  # e_uom.epUNIT_MW_m3
        btu_hft3 = 565  # e_uom.epUNIT_btu_hft3
        W_ft3 = 566  # e_uom.epUNIT_W_ft3
        W_kg = 567  # e_uom.epUNIT_W_kg
        ms = 568  # e_uom.epUNIT_ms
        mus = 569  # e_uom.epUNIT_mus
        Promille = 570  # e_uom.epUNIT_Promille
        Promille_x = 571  # e_uom.epUNIT_Promille_x
        Promille_xg = 572  # e_uom.epUNIT_Promille_xg
        Promille_vol = 573  # e_uom.epUNIT_Promille_vol
        Promille_kg = 574  # e_uom.epUNIT_Promille_kg
        Promille_x_rg = 575  # e_uom.epUNIT_Promille_x_rg
        mol_promille = 576  # e_uom.epUNIT_mol_promille
        mass_promille = 577  # e_uom.epUNIT_mass_promille
        unit_1_W = 578  # e_uom.epUNIT_1_W
        unit_1_V = 579  # e_uom.epUNIT_1_V
    except AttributeError:
        pass

    @staticmethod
    def str2unit(string: str):
        """
        将常见的单位字符串转换为EbsUnits对象

        :param string:
        :return:
        """
        string = string.replace("/", "_")
        string = string.replace("℃", "GrdC")
        return eval(f"EbsUnits.{string}")


class EbsResult:
    def __init__(self, result=None, result_code=-1):
        self.result = result
        if self.result is not None:
            self.result_code = self.result[0]
        else:
            self.result_code = result_code

    def success(self, no_error_as_success=False):
        """
        计算结果是否成功

        :param no_error_as_success: 没有错误则认为成功，相当于只看计算结果中是否有错误，没有则认为成功，这种情况下，达到最大迭代步数也会认为成功
        :return:
        """
        if no_error_as_success:
            success_list = [0, 1, 2, 6, 7]
        else:
            success_list = [0, 1, 2]
        if self.result_code in success_list:
            return True
        else:
            return False

    def get_result_detail(self):
        """
        打印并返回计算结果的所有信息，会逐条返回每一条error或warning或comment信息。

        :return:
        """
        if self.result is None:
            logger.debug("结果中不包括详细信息，如需详细信息，请使用model.simulate()或model.simulate2()方法！")
            msg = self.get_result_summary()
            logger.debug(f"当前计算结果为：{msg}")
            return {"结果总结": msg}
        else:
            result_dict = {"结果总结": self.get_result_summary()}
            results = self.result[1]
            for i in range(results.Count):
                idx = i + 1
                item = results.Item(idx)
                description = item.Description
                component_name = item.Object.Name
                result_dict.update({
                    idx: {
                        "description": description,
                        "component": component_name,
                        "level": item.Level,
                    }
                })
                logger.debug(f"组件名：{component_name:^30s}消息：{description}")  # 格式含义，^居中，20字符串宽度，s表示字符串
            return result_dict

    def get_result_summary(self):
        """
        获取计算结果是否成功的提示信息，即计算结果提示框的信息，这并不是具体的错误信息。

        :return:
        """
        # Function to translate EBSILON Calculation Result Codes to text
        ebs_err_code_txt = "Unknown Error Code"
        if self.result_code == 0:
            ebs_err_code_txt = "Simulation Successful"
        if self.result_code == 1:
            ebs_err_code_txt = "Simulation Successful With Comments"
        if self.result_code == 2:
            ebs_err_code_txt = "Simulation Successful With Warnings"
        if self.result_code == 3:
            ebs_err_code_txt = "Simulation Failed With Errors"
        if self.result_code == 4:
            ebs_err_code_txt = "Simulation Failed with Errors before Calculation - Check Model set up"
        if self.result_code == 5:
            ebs_err_code_txt = "Simulation Failed - Fatal Error"
        if self.result_code == 6:
            ebs_err_code_txt = "Simulation Failed - Maximum Number of Iterations Reached"
        if self.result_code == 7:
            ebs_err_code_txt = "Simulation Failed - Maximum Number of Iterations Reached With Warnings"
        if self.result_code == 8:
            ebs_err_code_txt = "Simulation Failed - Maximum Simulation Duration Time Exceeded"
        if self.result_code == 9:
            ebs_err_code_txt = "Simulation Failed - Maximum Number of Iterations Reached With Errors"
        if self.result_code == 10:
            ebs_err_code_txt = "Simulation Failed - License Error"
        if self.result_code == 11:
            ebs_err_code_txt = "Simulation Failed- Already In Simulation Error"
        if self.result_code == 12:
            ebs_err_code_txt = "Simulation Failed - Internal Error"
        return ebs_err_code_txt


class EbsApp:
    def __init__(self, ebsilon=None):
        """
        Ebsilon软件对象，参见EbsOpen.Application帮助文件，为空则会自动创建该对象。

        :param ebsilon:
        """
        if ebsilon is None:
            self.app = None  # ebsilon内部的application类对象，代表了ebsilon软件本身
            self.e_uom: Optional[win32com.client.Constants] = None  # ebsilon的枚举常量集合
            # 初始化ebsilon，返回Ebsilon软件内置的Application类对象
            try:
                # app = win32com.client.dynamic.Dispatch("EbsOpen.Application") #late binding / dynamic dispatch
                ebsilon = win32com.client.gencache.EnsureDispatch(
                    "EbsOpen.Application")  # early binding / static dispatch
                self.e_uom = win32com.client.constants
            except:
                logger.debug("初始化Ebsilon时发生错误，请检查您的Ebsilon license")
                exit()

        self.app = ebsilon
        # self.about = ebsilon.AboutString
        self.com_class_id = ebsilon.COMCLSID
        self.com_progress_id = ebsilon.COMProgID
        self.computation_tool = ebsilon.ComputationTool
        self.configuration = ebsilon.Configuration
        self.ebsilon_professional_flag = ebsilon.EbsilonProfessionalFlag
        self.ebsilon_professional_license_values_layout2 = ebsilon.EbsilonProfessionalLicenseValuesLayout2
        self.ebsilon_professional_license_values_layout2_marshal_array = \
            ebsilon.EbsilonProfessionalLicenseValuesLayout2MarshalArrayAsVariant
        self.ebs_open_demo_mode = ebsilon.EbsOpenDemoMode
        # self.events = ebsilon.Events  # 初始化时报错，可能是个动态变量
        # self.intel_mkl_cnr_cbwr = ebsilon.Intel_MKL_CNR_CBWR
        self.is_dll = ebsilon.IsDll
        self.isProcess = ebsilon.IsProcess
        self.module_filename = ebsilon.ModuleFileName
        self.object_caster = ObjectCaster(ebsilon.ObjectCaster)
        self.position = ebsilon.Position
        self.version_string = ebsilon.ProductVersionString
        self.version_major = ebsilon.ProductVersionMajor
        self.version_minor = ebsilon.ProductVersionMinor
        self.version_serialization = ebsilon.SerializationVersion
        self.unit_converter = ebsilon.UnitConverter

    def open(self, filename, on_error_return_null: bool = True):
        model: EbsModel = EbsModel(self.app.Open(filename, on_error_return_null))
        if not model.exists():
            logger.debug(f"打开模型文件时出错，检查文件路径是否存在：{filename}，或尝试初始化EbsApp")
        return model

    def get_active_model(self):
        return EbsModel(self.app.ActiveModel)

    def get_models(self):
        models = []
        for m in self.app.Models:
            models.append(EbsModel(m))
        return models

    def show_window(self, show=None):
        if show is None:
            show = self.e_uom.epShowWindowShowNormal
        result: bool = self.app.ShowWindow(show)
        return result

    def get_application_string(self):
        return self.app.ApplicationString

    def describe(self):
        logger.info(self.get_application_string())

    def get_object_caster(self):
        return ObjectCaster(self.app.ObjectCaster)


class EbsModel:
    def __init__(self, model):
        """
        Ebsilon软件的模型类，考虑到动态更新成员变量和循环引用，将源对象的成员变量更改为对应的成员方法模式获取

        :param model: Ebsilon软件待封装的模型对象
        """
        self.model = model

    def set_value(self, component, abbr, value: str | float, unit: EbsUnits = None, save_flag=True):
        """
        设置模型中某个组件中的参数

        :param component: 组件名
        :param abbr: 参数名，缩写
        :param value: 参数数值
        :param unit: 单位，EbsUnits对象，默认使用参数现有的单位
        :param save_flag: 设置后是否保存模型，默认保存
        :return:
        """
        comp = self.object_by_context(component)  # 不能使用get_object
        try:
            value = float(value)
        except:
            pass
        comp.set_value(abbr=abbr, value=value, unit=unit, save_flag=save_flag)

    def get_value(self, component, abbr, unit=None):
        comp = self.object_by_context(component)
        return comp.get_value(abbr, unit)

    def get_active_profile(self):
        return EbsProfile(self.model.ActiveProfile)

    def get_compression_level(self):
        return self.model.CompressionLevel

    def get_compression_method(self):
        return self.model.CompressionMethod

    def has_calculation_equations(self):
        return self.model.HasCalculationEquations

    def get_calculation_equations(self):
        return self.model.CalculationEquations

    def has_calculation_errors(self):
        return self.model.HasCalculationErrors

    def get_calculation_errors(self):
        return self.model.CalculationErrors

    def has_validation_ebsilon_entries(self):
        return self.model.HasValidationEbsilonEntries

    def get_validation_ebsilon_entries(self):
        return self.model.ValidationEbsilonEntries

    def get_statistics(self):
        return self.model.Statistics

    def get_doa(self):
        return self.model.DOA

    def get_toa(self):
        return self.model.TOA

    def get_context(self):
        return self.model.Context

    def get_objects(self):
        """
        获取模型中的所有组件列表

        :return:
        """
        objects = []
        for obj in self.model.Objects:
            objects.append(EbsObject(obj))
        return objects

    def get_position(self):
        return self.model.Position

    def get_path(self):
        return self.model.Path

    def get_root_profile(self):
        return self.model.RootProfile

    def get_summary(self):
        return self.model.Summary

    def get_name(self):
        return str(self.model.Name)

    def get_configuration(self):
        return self.model.Configuration

    def get_events(self):
        return self.model.Events

    def get_app(self):  # 不能定义为成员变量self.app，否则和EbsApp的构造方法循环引用
        return EbsApp(self.model.Application)

    def activate(self):
        self.model.Activate()

    def activate_profile(self, name_or_id: "str or int"):
        """
        激活指定的Profile

        :param name_or_id:
        :return:
        """
        result: bool = self.model.ActivateProfile(name_or_id)
        return result

    def get_profile(self, name_or_id: "str or int" = None, waring=True):
        """
        获取模型中指定的Profile，如果不存在指定的profile，则返回None

        :param name_or_id:
        :param waring: 不存在时是否报警告
        :return:
        """
        if name_or_id is None:
            return EbsProfile(self.model.ActiveProfile)
        else:
            _tuple = self.model.GetProfile(name_or_id)
            if _tuple[0]:
                return EbsProfile(_tuple[1])
            else:
                if waring:
                    logger.warning(f"不存在指定的profile：{name_or_id}")
                return None

    def get_all_profiles(self):
        profiles = []
        i = 0
        _ = self.get_profile(i, waring=False)
        while _ is not None:
            profiles.append(_)
            i = i + 1
            _ = self.get_profile(i, waring=False)
        return profiles

    def get_position1(self, from_current_context: bool = True):
        return self.model.getPosition(from_current_context)

    def get_object(self, name: str, on_not_found_return_null=True):
        """
        根据组件名获取组件对象，该方法会获取到具体的组件对象，而不是统一的EbsObject类对象

        :param name:
        :param on_not_found_return_null:
        :return:
        """
        obj: EbsObject = self.object_by_context(name, on_not_found_return_null)
        if not obj.exists():  # 查找的组件不存在
            return None
        result = obj.cast_to_component()  # 将当前组件转换为对应的组件类对象
        return result

    def object_by_context(self, name: str, on_not_found_return_null=True):
        result: EbsObject = EbsObject(self.model.ObjectByContext(name, on_not_found_return_null))
        if not result.exists():
            logger.warning(f"不存在名为{name}的组件")
        return result

    def save(self):
        result: bool = self.model.Save()
        return result

    def save_as(self, filepath: str):
        """
        保存一个当前模型的副本，当前打开的模型变更为新保存的模型

        :param filepath:
        :return:
        """
        if not os.path.exists(os.path.dirname(filepath)):
            os.makedirs(os.path.dirname(filepath))
        result: bool = self.model.SaveAs(filepath)
        return result

    def save_copy_as(self, filepath: str):
        """
        保存一个当前模型的副本，当前打开的模型不变
        :param filepath:
        :return:
        """
        result: bool = self.model.SaveCopyAs(filepath)
        return result

    def simulate(self):
        """
        对当前活动profile进行计算，计算返回提示信息，返回的提示信息过于丰富，不推荐使用，推荐使用simulate_new()方法
        :return:
        """
        result = None
        result = self.model.Simulate(result)  # Simulate方法需要传入形参
        result = EbsResult(result)
        return result

    def simulate2(self):
        """
        提示信息更丰富，会提示过时的设置，且提示信息比图形界面还丰富，不推荐使用

        :return:
        """
        result = None
        results = self.model.Simulate2(result)  # Simulate2方法需要传入形参
        results = EbsResult(results)
        return results

    def simulate_new(self):
        result = self.model.SimulateNew()  # SimulateNew方法不需要传入形参
        result = EbsResult(result_code=result)
        return result

    def describe(self, print_objects=False):
        logger.info(f"模型路径：{self.get_path()}")
        profiles = self.get_all_profiles()
        logger.info("该模型包含以下Profiles：")
        for prof in profiles:
            logger.info(f"id: {prof.get_profile_id()}，Name: {prof.get_name()}")
        if print_objects:
            objects = self.get_objects()
            objects = sorted(objects, key=lambda o: o.get_kind(), reverse=True)
            logger.info(f"该模型共包含{len(objects)}个组件，如下所示：")
            logger.info(f"{'name':^35}|{'type':^10}")
            for obj in objects:
                logger.info(f"{obj.get_name():^35}|{obj.get_kind():^10}")

    def exists(self):
        """
        当前模型是否存在

        :return:
        """
        if self.model is None:
            return False
        else:
            return True


class EbsProfile:
    def __init__(self, profile):
        """
        Ebsilon软件的Profile类，其中，考虑到循环引用和动态更新的问题，将成员变量更改为相应的成员方法
        :param profile:
        """
        self.profile = profile

    def get_configuration(self):
        return self.profile.Configuration

    def has_new_nominal_values(self):
        return self.profile.HasNewNominalValues

    def has_parent(self):
        return bool(self.profile.HasParent)

    def get_profile_id(self):
        return self.profile.ProfileId

    def is_active(self):
        return bool(self.profile.IsActive)

    def get_name(self):
        return str(self.profile.Name)

    def get_parent(self):
        # 不能定义为成员变量self.parent，否则和EbsApp类的构造方法循环引用
        return EbsProfile(self.profile.Parent)

    def get_app(self):
        # 不能定义为成员变量self.app，否则和EbsApp类的构造方法循环引用
        return EbsApp(self.profile.Application)

    def get_model(self):
        # 不能定义为成员变量，否则和EbsModel类的构造方法循环引用
        return EbsModel(self.profile.Model)

    def activate(self):
        result: bool = self.profile.Activate()
        return result

    def change_name(self, new_name: str):
        result: str = self.profile.ChangeName(new_name, allow_modify=True, check_only=False)
        return result

    def get_children(self):
        children = self.profile.Children
        result = []
        for child in children:
            result.append(EbsProfile(child))
        return result

    def copy(self, copied_profile, deep=True, parent_of_copied_profile=None):
        """
        方法未调试

        :param copied_profile:
        :param deep:
        :param parent_of_copied_profile:
        :return:
        """
        result: bool = self.profile.Copy(copied_profile, deep, parent_of_copied_profile)
        return result

    def delete(self):
        result: bool = self.profile.Delete()
        return result

    def make_to_root(self):
        result: bool = self.profile.MakeToRoot()
        return result

    def new_child(self):
        result: EbsProfile = EbsProfile(self.profile.NewChild())
        return result

    def saturate(self, profile: Optional["EbsProfile"]):
        """
        可能是全屏操作

        :param profile:
        :return:
        """
        result: bool = self.profile.Saturate(profile)
        return result

    def take_over_nominal_values(self):
        result: bool = self.profile.TakeOverNominalValues()
        return result


class ObjectCaster:
    def __init__(self, caster):
        self.caster = caster

    def cast_to_component(self, component_id: int, obj):
        """
        将IObject类型转换为具体的类型
        :param component_id: 因为Ebsilon有157个不同类型的组件，因此component_id≥1，≤157
        :return:
        """
        _ = self.caster
        obj = obj.obj
        expression = f"_.CastToComp{component_id}(obj)"
        result = eval(expression)
        return result

    def cast_to_data(self, obj):
        """
        将EbsObject类型转换为数据包类型，便于进一步提取组件中的数据

        :param obj:
        :return:
        """
        if isinstance(obj, EbsObject):
            obj = obj.obj
        result = self.caster.CastToData(obj)
        return result


class EbsObject:
    def __init__(self, obj):
        """
        Ebsilon软件的Object类

        :param obj: Ebsilon软件的Object类对象
        """
        self.obj = obj

    def get_data_object(self):
        return self.obj.DataObject

    def get_description(self):
        return self.obj.Description

    def get_description2(self):
        return self.obj.Description2

    def get_description3(self):
        return self.obj.Description3

    def get_description4(self):
        return self.obj.Description4

    def get_name(self):
        return self.obj.Name

    def get_fullname(self):
        return self.obj.FullName

    def get_font(self):
        return self.obj.Font

    def get_font_color(self):
        return self.obj.FontColor

    def get_model(self):
        return EbsModel(self.obj.Model)

    def get_app(self):
        return self.get_model().get_app()

    def get_kind(self):
        return self.obj.Kind

    def get_组件号(self, warning=True):
        组件号 = self.get_kind() - 10000
        if 1 <= 组件号 <= 157:
            return 组件号
        else:
            if warning:
                logger.warning(f"未知组件号：{self.get_kind()}")
                return -1
            else:
                return 组件号 + 10000

    def cast_to_object(self):
        return self.obj.CastToObject()

    def cast_to_component(self):
        """
        将Object对象转换为具体的组件类对象，如果是文本、管道等非组件类对象，则原样返回

        :return:
        """
        组件号 = self.get_组件号()
        if 组件号 == -1:
            return self
        ebs_app = self.get_model().get_app()
        caster = ebs_app.object_caster
        result = caster.cast_to_component(组件号, self)  # 将当前组件转换为对应的组件类对象
        return result

    def change_name(self, new_name: str, allow_modify=True, check_only=False):
        new_name: str = self.obj.ChangeName(new_name, allow_modify, check_only)
        return new_name

    def exists(self):
        if self.obj is None:
            return False
        else:
            return True

    def get_values(self):
        """
        获取部件所有参数的名称、数值和单位

        :return:
        """
        data_obj = self.get_app().get_object_caster().cast_to_data(self)
        parameters = data_obj.EbsValues
        values = []
        for para in parameters:
            values.append((para.Name, para.Value, EbsUnits(para.Unit).name))
        return values

    def get_value(self, abbr: str, unit: EbsUnits = None):
        """
        获取指定参数的值

        :param abbr: 参数名称缩写
        :param unit: 参数的单位，默认为参数当前单位，如需其他单位，传入该参数即可
        :return:
        """
        data_obj = self.get_app().get_object_caster().cast_to_data(self)
        value = data_obj.EbsValues(abbr)
        if value is None:
            logger.warning(f"未知参数名：{self.get_name()}.{abbr}")
            return None
        if unit is None:
            unit = EbsUnits(value.GetUnitValue()[1].Unit)
        if type(unit) == str:
            unit = EbsUnits.str2unit(str(unit))
        value = value.GetValueInUnit(unit.value)
        return value

    def get_abbr_unit(self, abbr) -> EbsUnits:
        """
        获取参数的当前单位

        :param abbr:
        :return: 返回EbsUnits对象
        """
        data_obj = self.get_app().get_object_caster().cast_to_data(self)
        value = data_obj.EbsValues(abbr)
        return EbsUnits(value.GetUnitValue()[1].Unit)

    def set_value(self, abbr: str, value, unit: EbsUnits = None, save_flag=True):
        """
        设置指定参数的值

        :param abbr: 参数名，在Ebsilon中为缩写
        :param value: 参数树值
        :param unit: 默认使用参数现有的单位
        :param save_flag: 为真则设置参数值后保存模型
        :return:
        """
        data_obj = self.get_app().get_object_caster().cast_to_data(self)  # 将当前组件转换为数据对象
        try:
            para = data_obj.EbsValues(abbr)
        except AttributeError:
            logger.error(abbr)
            return
        if para is None:
            logger.warning(f"未知参数名：{self.get_name()}.{abbr}")
            return None
        if unit is None:
            unit = EbsUnits(para.GetUnitValue()[1].Unit)  # 该语句可以查看当前变量的单位

        if is_type(unit, "str"):
            try:
                unit = EbsUnits.str2unit(str(unit))
            except:
                logger.debug(f"未知的单位：{unit}，只设置{self.get_name()}.{abbr}数值，忽略单位")
                unit = EbsUnits(para.GetUnitValue()[1].Unit)  # 该语句可以查看当前变量的单位
        bret_val = para.SetValueInUnit(value, unit.value)
        if bret_val and save_flag:
            self.get_model().save()

        return bret_val


def is_type(obj, type_str: str):
    actual_type_str = str(type(obj)).replace("'>", "")
    if actual_type_str.endswith(type_str):
        return True
    else:
        return False



from abc import ABCMeta, abstractmethod
import types
import warnings


class SkoBase(metaclass=ABCMeta):
    def register(self, operator_name, operator, *args, **kwargs):
        '''
        regeister udf to the class
        :param operator_name: string
        :param operator: a function, operator itself
        :param args: arg of operator
        :param kwargs: kwargs of operator
        :return:
        '''

        def operator_wapper(*wrapper_args):
            return operator(*(wrapper_args + args), **kwargs)

        setattr(self, operator_name, types.MethodType(operator_wapper, self))
        return self

    def fit(self, *args, **kwargs):
        warnings.warn('.fit() will be deprecated in the future. use .run() instead.'
                      , DeprecationWarning)
        return self.run(*args, **kwargs)


class Problem(object):
    pass



import os
import re
import traceback

import pandas as pd
from yangke.common.QtImporter import QColor, QFileDialog, QApplication

from yangke.base import dynamic_for
from yangke.common.config import logger
from yangke.ebsilon.ebsilon import EbsApp, EbsModel
from yangke.common.fileOperate import write_line
from yangke.common.qt import YkWindow, run_app, YkConsole


class MainWindow(YkWindow):
    def __init__(self):
        super().__init__()
        panel_ui_file = os.path.join(os.path.dirname(__file__), "UI", "ui_panel_ebsilon.yaml")
        table_ui_file = os.path.join(os.path.dirname(__file__), "UI", "ui_table_ebsilon.yaml")
        self.enable_input_panel(panel_ui_file, force=True)
        self.enable_table(table_ui_file, force=True)
        self.file_model: str | None = self.proj.get("file_model")
        self.file_summary: str | None = self.proj.get("file_summary")
        self.start_from = self.proj.get("start_from")
        self.err_folder = self.proj.get("err_folder")
        self.profile = self.proj.get("profile")
        self.cwd = self.proj.get("cwd")
        self._input_panel.set_value("Ebsilon模型", self.file_model)
        self._input_panel.set_value("结果保存至", self.file_summary)
        self._input_panel.set_value("开始计算序号", self.start_from)
        self._input_panel.set_value("错误模型保存目录", self.err_folder)
        self._input_panel.set_value("变工况Profile名", self.profile)
        self._table_widget.display_dataframe(self.proj.get("input_data"), row_index=2, col_index=0, header=None,
                                             index=None)
        self._table_widget.set_cell_value(1, 3, "参数描述")
        self._table_widget.item(1, 3).setBackground(QColor("#dfc2ea"))
        self._table_widget.display_dataframe(self.proj.get("output_data"), row_index=2, col_index=5, header=None,
                                             index=None)
        self.model: EbsModel | None = None
        self.ebs: EbsApp | None = None
        self.console = YkConsole()
        self.add_content_tab(self.console, "终端", True)
        logger.add(self.console)

    def choose_ebs_model(self):
        self.file_model, _ = QFileDialog.getOpenFileName(self, "打开项目", self.cwd,
                                                         "项目文件(*.ebs);;所有文件(*)")
        if self.file_model is not None:
            self._input_panel.set_value("Ebsilon模型", self.file_model)
            self.proj["file_model"] = self.file_model
            self.cwd = os.path.dirname(self.file_model)

    def choose_summary_file(self):
        self.file_summary, _ = QFileDialog.getSaveFileName(self, '保存项目', self.cwd,
                                                           "项目文件(*.csv);;所有文件(*)")
        if self.file_summary:
            self.proj["file_summary"] = self.file_summary
            self._input_panel.set_value("结果保存至", self.file_summary)

    def choose_error_folder(self):
        self.err_folder = QFileDialog.getExistingDirectory(self, '选择错误工况模型保存目录', self.cwd,
                                                           QFileDialog.ShowDirsOnly | QFileDialog.DontResolveSymlinks)
        if self.err_folder:
            self.proj["err_folder"] = self.err_folder
            self._input_panel.set_value("错误模型保存目录", self.err_folder)

    def apply(self):
        _ = self._input_panel.get_values_and_units(need_unit=False, need_dict=True)
        self.file_summary = _.get("结果保存至")
        self.file_model = _.get("Ebsilon模型")
        self.err_folder = _.get("错误模型保存目录")
        self.profile = _.get("变工况Profile名")
        self.start_from = float(_.get("开始计算序号"))
        self.proj.update({
            "file_summary": self.file_summary,
            "file_model": self.file_model,
            "err_folder": self.err_folder,
            "start_from": self.start_from,
            "profile": self.profile
        })
        self.read_input_output_para()

    def read_input_output_para(self):
        input_data = self._table_widget.read_data_in_range(2, 30, 1, 4, with_column=True)
        output_data = self._table_widget.read_data_in_range(2, 30, 6, 8, with_column=True)

        self.proj["input_data"] = input_data
        self.proj["output_data"] = output_data
        if self.proj_file is not None:
            self.save()

    def start_batch(self):
        self.apply()
        self._content_tab.activate_tab("终端")
        self.ebs = EbsApp()
        self.ebs.describe()
        self.batch_all_conditions()

    def batch_all_conditions(self):
        self.read_model()  # 读入ebsilon模型文件
        model = self.model
        model.activate_profile(self.profile)  # 切换到专为批量计算设计的batch工况

        i = 0
        input_data: pd.DataFrame = self.proj.get("input_data")
        # 删除input_data中以=开头的行，这些行是参数设置，不从循环中设置
        # 将input_data转换为list类型的二位数组
        data_for = []
        data_set_value = []
        components = []
        des_input = []
        paras = []
        _ = input_data.T
        for idx, row in _.items():  # .items()方法只能按列遍历，因此需要先转置
            if row["参数取值"].startswith("="):
                data_set_value.append(row)
            else:
                data_for.append(eval(row["参数取值"]))
                components.append(row["组件名"])
                paras.append(row["参数名"])
                des_input.append(row["参数描述"])

        # ------------------------------------- 写输出文件的标题行 -------------------------------------
        res_para_description = []
        _ = self.proj.get("output_data").T
        for idx, row in _.items():
            res_para_description.append(row["参数描述"])
        line = ["序号", *des_input, *res_para_description, "计算结果"]
        line = ",".join(line)
        write_line(self.file_summary, f"{line}\n")
        # ------------------------------------- 写输出文件的标题行 -------------------------------------

        # ------------------------------------- 剔除不满足限制条件的工况 -------------------------------------

        # ------------------------------------- 剔除不满足限制条件的工况 -------------------------------------


        for input_values in dynamic_for(data_for):
            self.process_events()
            if i < self.start_from:
                i = i + 1
                continue
            logger.debug(f"=========================={i} start===========================")
            logger.debug(f"输入参数为：{des_input} = {input_values}")
            # 循环设置每个参数
            for com, par, dat in zip(components, paras, input_values):
                try:
                    try:
                        dat = float(dat)
                    except:
                        pass
                    model.set_value(com, par, dat)
                except (AttributeError, NameError):
                    logger.error(f"设置模型的值时发生错误")
                    traceback.print_exc()
                    return

            # 循环设置条件参数，有些参数的取值为条件语句，如 当Comp_1.Q<=0时，com.par取1，否则取2，则dat="=1 if Comp_1.Q<=0 else 2"
            for row in data_set_value:
                com, par, dat, _ = row.values
                res = re.findall(r"=(.+) if (.+) else (.+)", dat)[0]  # dat = "=1 if Comp_1.Q<=0 else 2"
                if len(res) == 3:
                    _com_prefer, _par_refer, _op, _val = re.findall(r"(.+?)\.(.+?)([<=>]+)(.+?)", res[1])[0]
                    _val_act = model.get_value(_com_prefer, _par_refer)
                    expression = f"{_val_act}{_op}{_val}"
                    if eval(expression):
                        val = res[0]
                    else:
                        val = res[2]

                model.set_value(com, par, val)

            result = model.simulate_new()
            cal_res_description = result.get_result_summary()
            logger.debug(cal_res_description)

            if result.success(no_error_as_success=True):
                # 计算成功，则组装计算结果
                output_values = []
                _ = self.proj.get("output_data").T
                for idx, row in _.items():
                    com = row["组件名"]
                    par = row["参数名"]
                    des = row["参数描述"]

                    if par.startswith("="):
                        # 说明结果是公式
                        val = self.evaluate_equation(par)
                    else:
                        val = model.get_value(com, par)
                    output_values.append(val)

                logger.debug(f"输出参数为：{res_para_description} = {output_values}")
                output_line = [i, *input_values, *output_values, cal_res_description]
            else:
                # 输出错误信息
                self.model.save_as(os.path.join(self.err_folder,
                                                f"model{i}.ebs"))
                output_line = [i, *input_values, *[None] * len(res_para_description), cal_res_description]

            output_line = [str(_) for _ in output_line]
            output_line = ",".join(output_line)
            write_line(self.file_summary, line=f"{output_line}\n", append=True)
            logger.debug(f"=========================={i} end===========================")
            i = i + 1

    def process_events(self):
        QApplication.processEvents()

    def evaluate_equation(self, equations: str):
        """
        根据字符串表达的公式计算公式的值
        """
        # ------------------------------- 分析公式 ---------------------------------
        if equations.startswith("="):
            equations = equations[1:]
        reg1 = "[A-Za-z]+[A-Za-z\d_]*\."  # 匹配以字母开头，后续跟任意数量的[字母数字下划线字符]，然后接一个小数点
        reg2 = "[A-Za-z]+[A-Za-z\d_]*"  # 匹配以字母开头，后续跟任意数量的[字母数字下划线字符]
        reg = f"({reg1}{reg2})"  # 组合后，即匹配 组件名.变量名 的字符串结果
        res = re.findall(reg, equations)  # 可能会有重复的，但不影响结果，去重操作计算量可能大于直接替换，因为大部分情况无重复
        for para in res:
            com, par = para.split('.')
            val = self.model.get_value(com, par)
            equations = equations.replace(para, str(val))
        _ = eval(equations)
        return _

    def read_model(self):
        model_path = self.file_model
        try:
            if self.model is not None and self.model.get_path() == model_path and self.model.model is not None:
                pass
            else:
                self.model = self.ebs.open(model_path)
        except AttributeError:  # 有时候会出现model不为空，但model.model为None的情况，这种情况下，model.get_path()会报错
            self.model = self.ebs.open(model_path)
        if self.model.model is None:
            self.ebs = EbsApp()  # 出现这种情况一般是ebs崩了，尝试重新初始化
            self.read_model()
        return self.model

    def get_color(self):
        ...


if __name__ == "__main__":
    # settings = server.read_settings(r"C:\Users\YangKe\Desktop\新建文件夹\15版\settings.xlsx")
    # ebs = EbsApp()
    # ebs.describe()
    # batch_all_conditions(start_from=0)
    run_app(MainWindow)



flow_全厂补水 = 100
flow_三期供热回水 = 10000
flow_四期供热回水 = 11000
flow_5号供热 = 4500
flow_6号供热 = flow_三期供热回水 - flow_5号供热 + flow_全厂补水 / 2
flow_7号供热 = 6000
flow_8号供热 = flow_四期供热回水 - flow_7号供热 + flow_全厂补水 / 2
flow_全厂供水 = flow_5号供热 + flow_6号供热 + flow_7号供热 + flow_8号供热

t_供热计划温度 = 90

t_modify_供热回水 = 0
t_dif_三四期 = 2

# 机组额定出力
power_design = [300, 300, 300, 300]
# 调峰补贴负荷上限，按照辅助调峰服务考核细则，调峰报价仅在低于70%负荷率的情况下填报，70%负荷以上的成本价格可用于分析调峰分摊的盈亏情况。
power_补贴上限 = [p * 0.7 for p in power_design]

# 是否供热
heat_supply_no5 = True
heat_supply_no6 = True
heat_supply_no7 = True
heat_supply_no8 = True

三期供热机组台数 = 0
四期供热机组台数 = 0
供热机组台数 = 0
if heat_supply_no5:
    供热机组台数 = 供热机组台数 + 1
    三期供热机组台数 = 三期供热机组台数 + 1
if heat_supply_no6:
    供热机组台数 = 供热机组台数 + 1
    三期供热机组台数 = 三期供热机组台数 + 1
if heat_supply_no7:
    供热机组台数 = 供热机组台数 + 1
    四期供热机组台数 = 四期供热机组台数 + 1
if heat_supply_no8:
    供热机组台数 = 供热机组台数 + 1
    四期供热机组台数 = 四期供热机组台数 + 1

# 是否切缸
cut_no5 = True
cut_no6 = True
cut_no7 = False
cut_no8 = True

t_供热回水 = 45 + (t_供热计划温度 - 80) / 2 + t_modify_供热回水
供热负荷_全厂 = flow_全厂供水 / 3.6 * (t_供热计划温度 - t_供热回水) * 4.19 / 1000

t_no5_供热回水 = t_供热回水 + t_dif_三四期 / 2
t_no6_供热回水 = t_供热回水 + t_dif_三四期 / 2
t_no7_供热回水 = t_供热回水 - t_dif_三四期 / 2
t_no8_供热回水 = t_供热回水 - t_dif_三四期 / 2
三期偏置 = 0

五号偏置 = 10
七号偏置 = 0

供热负荷_no5 = 0
供热负荷_no6 = 0
供热负荷_no7 = 0
供热负荷_no8 = 0
if heat_supply_no5:
    供热负荷_no5 = 供热负荷_全厂 / 供热机组台数 + 三期偏置 * 4.2
    供热负荷_no5 = 供热负荷_no5 + 五号偏置 * 4.2 if heat_supply_no6 else 供热负荷_no5
if heat_supply_no6:
    供热负荷_no6 = 供热负荷_全厂 / 供热机组台数 + 三期偏置 * 4.2
    供热负荷_no6 = 供热负荷_no6 - 五号偏置 * 4.2 if heat_supply_no5 else 供热负荷_no6
if heat_supply_no7:  # 在该情况下，四期供热机组台数必然大于0
    供热负荷_no7 = 供热负荷_全厂 / 供热机组台数 - 三期偏置 * 4.2 * 三期供热机组台数 / 四期供热机组台数
    供热负荷_no7 = 供热负荷_no7 + 七号偏置 * 4.2 if heat_supply_no8 else 供热负荷_no7
if heat_supply_no8:
    供热负荷_no8 = 供热负荷_全厂 / 供热机组台数 - 三期偏置 * 4.2 * 三期供热机组台数 / 四期供热机组台数
    供热负荷_no8 = 供热负荷_no7 - 七号偏置 * 4.2 if heat_supply_no7 else 供热负荷_no8

t_no5_供热 = 供热负荷_no5 * 1000 / flow_5号供热 * 3.6 / 4.19 + t_no5_供热回水
t_no6_供热 = 供热负荷_no6 * 1000 / flow_6号供热 * 3.6 / 4.19 + t_no6_供热回水
t_no7_供热 = 供热负荷_no7 * 1000 / flow_7号供热 * 3.6 / 4.19 + t_no7_供热回水
t_no8_供热 = 供热负荷_no8 * 1000 / flow_8号供热 * 3.6 / 4.19 + t_no8_供热回水

energy_no5_供热 = flow_5号供热 * (t_no5_供热 - t_no5_供热回水) * 4.19 * 24 / 1000
energy_no6_供热 = flow_5号供热 * (t_no6_供热 - t_no5_供热回水) * 4.19 * 24 / 1000
energy_no7_供热 = flow_5号供热 * (t_no7_供热 - t_no5_供热回水) * 4.19 * 24 / 1000
energy_no8_供热 = flow_5号供热 * (t_no8_供热 - t_no5_供热回水) * 4.19 * 24 / 1000

供热当量电负荷_no5 = 供热负荷_no5 / 4.42
供热当量电负荷_no6 = 供热负荷_no6 / 4.56
供热当量电负荷_no7 = 供热负荷_no7 / 4.5
供热当量电负荷_no8 = 供热负荷_no8 / 4.8
供热当量电负荷 = 供热当量电负荷_no5 + 供热当量电负荷_no6 + 供热当量电负荷_no7 + 供热当量电负荷_no8

电负荷_max_no5 = 320 - 供热当量电负荷_no5 if heat_supply_no5 else 300
电负荷_max_no6 = 320 - 供热当量电负荷_no6 if heat_supply_no6 else 300
电负荷_max_no7 = 320 - 供热当量电负荷_no7 if heat_supply_no7 else 300
电负荷_max_no8 = 320 - 供热当量电负荷_no8 if heat_supply_no8 else 300
电负荷_max = 电负荷_max_no5 + 电负荷_max_no6 + 电负荷_max_no7 + 电负荷_max_no8


def 下限电负荷(is_供热, is_cut, 供热当量电负荷_local=0):
    if is_供热:
        if is_cut:
            result = max(供热当量电负荷_local * 2, 150 - 供热当量电负荷_local)
        else:
            result = max(供热当量电负荷_local * 1.65 + 70, 120 - 供热当量电负荷_local)
    else:
        result = 120
    return result


电负荷_min_no5 = 下限电负荷(heat_supply_no5, cut_no5, 供热当量电负荷_no5)
电负荷_min_no6 = 下限电负荷(heat_supply_no6, cut_no6, 供热当量电负荷_no6)
电负荷_min_no7 = 下限电负荷(heat_supply_no7, cut_no7, 供热当量电负荷_no7)
电负荷_min_no8 = 下限电负荷(heat_supply_no8, cut_no8, 供热当量电负荷_no8)

电负荷_min = 电负荷_min_no5 + 电负荷_min_no6 + 电负荷_min_no7 + 电负荷_min_no8

price_coal = 806
price_elec = 0.33
price_heat = 25.69
考虑电量比例 = 0
单位煤耗变动 = 2.74

效率成本变化 = price_coal * 单位煤耗变动 / 1000 + price_elec
上限供热成本 = 10000 / 65 * (price_heat - 20 * price_coal / 1000 - 3 * price_elec) / 10
下限供热成本 = 3.6 * 10 * 2.5 * (price_heat - 20 * price_coal / 1000 - 3 * price_elec) / 10

电负荷_list = range(80, 301, 10)
